#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import re
from decimal import *

from ikabot.config import *

getcontext().prec = 30


def getAvailableResources(html, num=False):
    """
    Parameters
    ----------
    html : string

    Returns
    -------
    resources_available : list[int] | list[str]
    """
    resources = re.search(
        r'\\"resource\\":(\d+),\\"2\\":(\d+),\\"1\\":(\d+),\\"4\\":(\d+),\\"3\\":(\d+)}',
        html,
    )
    if num:
        return [
            int(resources.group(1)),
            int(resources.group(3)),
            int(resources.group(2)),
            int(resources.group(5)),
            int(resources.group(4)),
        ]
    else:
        return [
            resources.group(1),
            resources.group(3),
            resources.group(2),
            resources.group(5),
            resources.group(4),
        ]


def getWarehouseCapacity(html):
    """
    Parameters
    ----------
    html : string
    Returns
    -------
    capacity : int
    """
    capacity = re.search(
        r'maxResources:\s*JSON\.parse\(\'{\\"resource\\":(\d+),', html
    ).group(1)
    return int(capacity)


def getWineConsumptionPerHour(html):
    """
    Parameters
    ----------
    html : string
    Returns
    -------
    capacity : int
    """
    result = re.search(r"wineSpendings:\s(\d+)", html)
    if result:
        return int(result.group(1))
    return 0


def getProductionPerHour(session, city_id):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_id : int

    Returns
    -------
    production: tuple[Decimal, Decimal, int]
    """
    resource_search_pool = {
        1: "js_GlobalMenu_production_wine",
        2: "js_GlobalMenu_production_marble",
        3: "js_GlobalMenu_production_crystal",
        4: "js_GlobalMenu_production_sulfur",
    }

    prod = session.get('?view=city&cityId=' + city_id)
    luxury_type_match = re.search(r'tradegood&type=(\d+)', prod)
    if not luxury_type_match:
        raise ValueError(f"Could not determine luxury resource type for city {city_id}")
    luxury_type = int(luxury_type_match.group(1))

    # The group ([\d,\s]+) now matches digits, commas, AND spaces.
    production_pattern = r'<td id="{}"[^>]*>\s*([\d,\s]+)\s*</td>'

    # Get wood production
    wood_match = re.search(production_pattern.format("js_GlobalMenu_resourceProduction"), prod)
    # Get luxury production
    luxury_match = re.search(production_pattern.format(resource_search_pool[luxury_type]), prod)

    if not wood_match or not luxury_match:
        raise AttributeError("Could not find production values. The game's HTML may have changed.")
    
    # helper function to strip all non-digit characters.
    def clean_number(num_str):
        return re.sub(r'[^\d]', '', num_str)

    # Use the helper to clean the strings before converting to int.
    wood_prod = int(clean_number(wood_match.group(1)))
    luxury_prod = int(clean_number(luxury_match.group(1)))

    wood_production = Decimal(wood_prod)
    luxury_production = Decimal(luxury_prod)
    luxury_resource_type = int(luxury_type)

    return wood_production, luxury_production, luxury_resource_type


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import math
import re
import sys
from decimal import *

from ikabot import config
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.naval import *
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.varios import *
from ikabot.helpers.pedirInfo import getShipCapacity


def isHostile(movement):
    """
    Parameters
    ----------
    movement : dict

    Returns
    -------
    is hostile : bool
    """
    if movement["army"]["amount"]:
        return True
    for mov in movement["fleet"]["ships"]:
        if mov["cssClass"] != "ship_transport":
            return True
    return False


def shipMovements(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        print(
            "Ships {:d}/{:d}\n".format(
                getAvailableShips(session), getTotalShips(session)
            )
        )

        cityId = getCurrentCityId(session)
        url = "view=militaryAdvisor&oldView=city&oldBackgroundView=city&backgroundView=city&currentCityId={}&actionRequest={}&ajax=1".format(
            cityId, actionRequest
        )
        resp = session.post(url)
        resp = json.loads(resp, strict=False)
        movements = resp[1][1][2]["viewScriptParams"]["militaryAndFleetMovements"]
        time_now = int(resp[0][1]["time"])

        if len(movements) == 0:
            print("There are no movements")
            enter()
            event.set()
            return

        for movement in movements:

            color = ""
            if movement["isHostile"]:
                color = bcolors.RED + bcolors.BOLD
            elif movement["isOwnArmyOrFleet"]:
                color = bcolors.BLUE + bcolors.BOLD
            elif movement["isSameAlliance"]:
                color = bcolors.GREEN + bcolors.BOLD

            origin = "{} ({})".format(
                movement["origin"]["name"], movement["origin"]["avatarName"]
            )
            destination = "{} ({})".format(
                movement["target"]["name"], movement["target"]["avatarName"]
            )
            arrow = "<-" if movement["event"]["isFleetReturning"] else "->"
            time_left = int(movement["eventTime"]) - time_now
            print(
                "{}{} {} {}: {} ({}) {}".format(
                    color,
                    origin,
                    arrow,
                    destination,
                    movement["event"]["missionText"],
                    daysHoursMinutes(time_left),
                    bcolors.ENDC,
                )
            )

            if movement["isHostile"]:
                troops = movement["army"]["amount"]
                fleets = movement["fleet"]["amount"]
                print(
                    "Troops:{}\nFleets:{}".format(
                        addThousandSeparator(troops), addThousandSeparator(fleets)
                    )
                )
            elif isHostile(movement):
                troops = movement["army"]["amount"]
                ships = 0
                fleets = 0
                for mov in movement["fleet"]["ships"]:
                    if mov["cssClass"] == "ship_transport":
                        ships += int(mov["amount"])
                    else:
                        fleets += int(mov["amount"])
                print(
                    "Troops:{}\nFleets:{}\n Ships:{}".format(
                        addThousandSeparator(troops),
                        addThousandSeparator(fleets),
                        addThousandSeparator(ships),
                    )
                )
            else:
                assert len(materials_names) == 5
                total_load = 0
                for resource in movement["resources"]:
                    amount = resource["amount"]
                    tradegood = resource["cssClass"].split()[1]
                    # gold won't be translated
                    if tradegood != "gold":
                        index = materials_names_tec.index(tradegood)
                        tradegood = materials_names[index]
                    total_load += int(amount.replace(",", "").replace(".", ""))
                    print("{} of {}".format(amount, tradegood))
                ship_capacity, freighter_capacity = getShipCapacity(session)
                ships = int(math.ceil((Decimal(total_load) / Decimal(ship_capacity))))
                print("{:d} Ships".format(ships))
        enter()
        event.set()
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import re

from ikabot.config import *
from ikabot.helpers.getJson import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.resources import *
from ikabot.helpers.varios import *


def donate(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        city = chooseCity(session)
        banner()

        woodAvailable = city["availableResources"][0]

        islandId = city["islandId"]
        html = session.get(island_url + islandId)
        island = getIsland(html)

        island_type = island["tipo"]
        resource_name = tradegoods_names[0]
        tradegood_name = tradegoods_names[int(island_type)]

        # get resource information
        url = "view=resource&type=resource&islandId={0}&backgroundView=island&currentIslandId={0}&actionRequest={1}&ajax=1".format(
            islandId, actionRequest
        )
        resp = session.post(url)
        resp = json.loads(resp, strict=False)

        resourceLevel = resp[0][1]["backgroundData"]["resourceLevel"]
        tradegoodLevel = resp[0][1]["backgroundData"]["tradegoodLevel"]
        resourceEndUpgradeTime = int(
            resp[0][1]["backgroundData"]["resourceEndUpgradeTime"]
        )
        resourceUpgrading = resourceEndUpgradeTime > 0
        tradegoodEndUpgradeTime = int(
            resp[0][1]["backgroundData"]["tradegoodEndUpgradeTime"]
        )
        tradegoodUpgrading = tradegoodEndUpgradeTime > 0

        if resourceUpgrading:
            resourceUpgradeMsg = "(upgrading, ends in:{})".format(
                daysHoursMinutes(resourceEndUpgradeTime)
            )
        else:
            resourceUpgradeMsg = ""
        if tradegoodUpgrading:
            tradegoodUpgradeMsg = "(upgrading, ends in:{})".format(
                daysHoursMinutes(tradegoodEndUpgradeTime)
            )
        else:
            tradegoodUpgradeMsg = ""

        html = resp[1][1][1]
        wood_total_needed, wood_donated = re.findall(
            r'<li class="wood">(.*?)</li>', html
        )
        wood_total_needed = wood_total_needed.replace(",", "").replace(".", "")
        wood_total_needed = int(wood_total_needed)
        wood_donated = wood_donated.replace(",", "").replace(".", "")
        wood_donated = int(wood_donated)

        if resourceUpgrading and tradegoodUpgrading:
            print(
                
                "Both the {} (ends in:{}) and the {} (ends in:{}) are being upgraded rigth now.\n".format(
                    resource_name,
                    daysHoursMinutes(resourceEndUpgradeTime),
                    tradegood_name,
                    daysHoursMinutes(tradegoodEndUpgradeTime),
                )
                
            )
            enter()
            event.set()
            return

        print("{} lv:{} {}".format(resource_name, resourceLevel, resourceUpgradeMsg))
        print(
            "{} / {} ({}%)\n".format(
                addThousandSeparator(wood_donated),
                addThousandSeparator(wood_total_needed),
                addThousandSeparator(int((100 * wood_donated) / wood_total_needed)),
            )
        )

        # get tradegood information
        url = "view=tradegood&type={0}&islandId={1}&backgroundView=island&currentIslandId={1}&actionRequest={2}&ajax=1".format(
            island_type, islandId, actionRequest
        )
        resp = session.post(url)

        resp = json.loads(resp, strict=False)
        html = resp[1][1][1]
        tradegood_total_needed, tradegood_donated = re.findall(
            r'<li class="wood">(.*?)</li>', html
        )
        tradegood_total_needed = tradegood_total_needed.replace(",", "").replace(
            ".", ""
        )
        tradegood_total_needed = int(tradegood_total_needed)
        tradegood_donated = tradegood_donated.replace(",", "").replace(".", "")
        tradegood_donated = int(tradegood_donated)

        print("{} lv:{} {}".format(tradegood_name, tradegoodLevel, tradegoodUpgradeMsg))
        print(
            "{} / {} ({}%)\n".format(
                addThousandSeparator(tradegood_donated),
                addThousandSeparator(tradegood_total_needed),
                addThousandSeparator(
                    int((100 * tradegood_donated) / tradegood_total_needed)
                ),
            )
        )

        print("Wood available:{}\n".format(addThousandSeparator(woodAvailable)))

        if resourceUpgrading is False and tradegoodUpgrading is False:
            msg = "Donate to {} (1) or {} (2)?:".format(
                resource_name, tradegood_name
            )
            donation_type = read(msg=msg, min=1, max=2)
            name = resource_name if donation_type == 1 else tradegood_name
            print("")
        else:
            if resourceUpgrading is False and tradegoodUpgrading is True:
                donation_type = 1
                name = resource_name
            else:
                donation_type = 2
                name = tradegood_name
            print("Donate to:{}\n".format(name))

        donation_type = "resource" if donation_type == 1 else "tradegood"

        amount = read(
            min=0,
            max=woodAvailable,
            default=0,
            additionalValues=["all", "half"],
            msg="Amount (number, all, half):",
        )
        if amount == "all":
            amount = woodAvailable
        elif amount == "half":
            amount = woodAvailable // 2
        elif amount == 0:
            event.set()
            return
        print("Will donate {} to the {}?".format(addThousandSeparator(amount), name))
        print("\nProceed? [Y/n]")
        rta = read(values=["y", "Y", "n", "N", ""])
        if rta.lower() == "n":
            event.set()
            return

        # do the donation
        session.post(
            params={
                "islandId": islandId,
                "type": donation_type,
                "action": "IslandScreen",
                "function": "donate",
                "donation": amount,
                "backgroundView": "island",
                "templateView": "resource",
                "actionRequest": actionRequest,
                "ajax": "1",
            }
        )

        print("\nDonation successful.")
        enter()
        event.set()
        return
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import re
import time
from math import ceil, floor
from ikabot.helpers.resources import *
from ikabot.helpers.varios import decodeUnicodeEscape
from ikabot.config import *
from typing import Dict, TypedDict, Optional, Any
# TODO replace Optional with NotRequried when we drop support for python 3.9 and 3.10. NotRequired is more precise here than Optional
LinkDict = TypedDict("LinkDict", {"onclick": str, "href": str, "tooltip": str})

FlyingTraderDict = TypedDict("FlyingTraderDict", {"link": LinkDict, "banner": str})

VisibilityDict = TypedDict(
    "VisibilityDict",
    {
        "military": int,
        "espionage": int,
        "resourceShop": int,
        "slot1": int,
        "slot2": int,
        "slot3": int,
        "slot4": int,
    },
)

CityLeftMenuDict = TypedDict("CityLeftMenuDict", {"visibility": VisibilityDict, "ownCity": int})

PositionDict = TypedDict(
    "PositionDict",
    {
        "name": str,
        "level": Optional[int],
        "isBusy": bool,
        "canUpgrade": Optional[bool],
        "isMaxLevel": Optional[bool],
        "building": str,
        "shipIsAtDockyard": Optional[int],

        # Added for compatibility
        "position": int,
        "type": Optional[str],
    },
)

FullCityDict = TypedDict(
    "FullCityDict",
    {
        "name": str,
        "id": str,
        "phase": int,
        "isCapital": bool,
        "ownerId": str,
        "ownerName": str,
        "islandId": str,
        "islandName": str,
        "islandXCoord": str,
        "islandYCoord": str,
        "buildingSpeedupActive": int,
        "showPirateFortressBackground": int,
        "showPirateFortressShip": int,
        "underConstruction": int,
        "endUpgradeTime": int,
        "startUpgradeTime": int,
        "position": list[PositionDict],
        "beachboys": str,
        "spiesInside": None,
        "cityLeftMenu": CityLeftMenuDict,
        "walkers": list[Any],
        "displayStaticPlague": bool,
        "dailyTasks": str,
        "cityCinema": str,
        "flyingTrader": FlyingTraderDict,

        # Added for compatibility
        "x": str,
        "y": str,
        "cityName": str,
        "isOwnCity": bool,
        "availableResources": list[int],
        "storageCapacity": int,
        "freeCitizens": int,
        "wineConsumptionPerHour": int,
        "resourcesListedForSale": list[int],
        "freeSpaceForResources": list[int],
    },
)


AvatarScoreDict = TypedDict(
    "AvatarScoreDict",
    {
        "avatar_id": str,
        "place": str,
        "building_score_main": str,
        "research_score_main": str,
        "army_score_main": str,
        "trader_score_secondary": str,
    },
)

AvatarScoresDict = Dict[str, AvatarScoreDict]

BarbariansDict = TypedDict(
    "BarbariansDict",
    {
        "invisible": int,
        "actionTitle": str,
        "actionClass": str,
        "actionLink": str,
        "count": int,
        "wallLevel": int,
        "level": int,
        "underAttack": int,
        "isTradegoodSiege": int,
        "city": str,
        "destroyed": int,
    },
)

CityDict = TypedDict(
    "CityDict",
    {
        "type": str,
        "name": str,
        "id": int,
        "level": int,
        "ownerId": Optional[str],
        "ownerName": Optional[str],
        "ownerAllyId": Optional[int],
        "ownerAllyTag": Optional[str],
        "hasTreaties": Optional[int],
        "actions": Optional[list[Any]],
        "state": Optional[str],
        "viewAble": int,
        "infestedByPlague": Optional[bool],
        "buildplace_type": Optional[str],

        # Added for compatibility

        "Id": Optional[str],
        "Name": Optional[str],
        "AllyId": Optional[int],
        "AllyTag": Optional[str],
        "_type": Optional[str],
    },
)

IslandDict = TypedDict(
    "IslandDict",
    {
        "id": str,
        "type": int,
        "name": str,
        "xCoord": str,
        "yCoord": str,
        "tradegood": int,
        "tradegoodTarget": str,
        "resourceLevel": str,
        "tradegoodLevel": str,
        "wonder": str,
        "wonderLevel": str,
        "wonderName": str,
        "showResourceWorkers": int,
        "showTradegoodWorkers": int,
        "showAgora": int,
        "canEnterResource": int,
        "canEnterTradegood": int,
        "tradegoodEndUpgradeTime": int,
        "resourceEndUpgradeTime": int,
        "wonderEndUpgradeTime": int,
        "isOwnCityOnIsland": bool,
        "cities": list[CityDict],
        "barbarians": BarbariansDict,
        "avatarScores": AvatarScoresDict,
        "specialServerBadges": list[Any],
        "selectedCityParameters": list[Any],
        "island": int,
        "isHeliosTowerBuilt": bool,
        "heliosTop": int,
        "heliosMid": int,
        "heliosBase": int,
        "heliosName": str,
        "heliosTooltip": str,
        "heliosActive": int,
        "showResourceBonusIcon": int,
        "showTradegoodBonusIcon": int,
        "walkers": list[Any],

        # Added for compatibility

        "x": int,
        "y": int,
        "tipo": str,
    },
)

WorldMapIslandDict = TypedDict(
    "WorldMapIslandDict",
    {
        "x": int,
        "y": int,
        "id": int,
        "name": str,
        "resourceType": int,
        "miracleType": int,
        "unknownValue1": int,
        "unknownValue2": int,
        "woodLvl": int,
        "cityCount": int,
        "piracyInRange": bool,
        "heliosTower": bool,
        "red": bool,
        "blue": bool,
        
        # Added for ease of use
        "resourceName": str,
        "miracleName": str,
    },
)


def getFreeCitizens(html: str) -> int:
    """This function is used in the ``getCity`` function to determine the amount of free (idle) citizens in the given city.
    Parameters
    ----------
    html : str
        a string representing html which is returned when sending a get request to view a city.

    Returns
    -------
    freeCitizens : int
        an integer representing the amount of free citizens in the given city.
    """
    freeCitizens = re.search(r'js_GlobalMenu_citizens">(.*?)</span>', html).group(1)
    freeCitizens = re.sub(r'\D', '', freeCitizens)
    return int(freeCitizens)


def getResourcesListedForSale(html: str) -> list[int]:
    """This function is used in the ``getCity`` function to determine the amount of each resource which is listed for sale in the branch office
    Parameters
    ----------
    html : str
        a string representing html which is returned when sending a get request to view a city.

    Returns
    -------
    onSale : list[int]
        a list containing 5 integers each of which representing the amount of that particular resource which is on sale in the given city. For more information about the order of the resources, refer to ``config.py``
    """
    rta = re.search(
        r'branchOfficeResources: JSON\.parse\(\'{\\"resource\\":\\"(\d+)\\",\\"1\\":\\"(\d+)\\",\\"2\\":\\"(\d+)\\",\\"3\\":\\"(\d+)\\",\\"4\\":\\"(\d+)\\"}\'\)',
        html,
    )
    if rta:
        return [
            int(rta.group(1)),
            int(rta.group(2)),
            int(rta.group(3)),
            int(rta.group(4)),
            int(rta.group(5)),
        ]
    else:
        return [0, 0, 0, 0, 0]

def getWorldMapIslands(html: str) -> list[WorldMapIslandDict]:
    """This function uses the html passed to it as a string to extract, parse and return a list of WorldMapIsland objects
    Parameters
    ----------
    html : str
        the html or json returned when a get request is made to the `view=worldmap_iso` or `action=WorldMap&function=getJSONArea` endpoints respectively. 

    Returns
    -------
    islands : list[WorldMapIsland]
        this function returns a json parsed list of WorldMapIsland objects.
    """

    isla = re.search(r"jsonData = '([\S\s]*?)'", html).group(1) if '!DOCTYPE html' in html else html

    worldMapIslands = []

    data = json.loads(isla)['data']
    for x in data:
        for y in data[x]:
            worldMapIslands.append({
                'x': int(x),
                'y': int(y),
                'id': int(data[x][y][0]),
                'name': data[x][y][1],
                'resourceType': int(data[x][y][2]),
                'miracleType': int(data[x][y][3]),
                'unknownValue1': int(data[x][y][4]),
                'unknownValue2': int(data[x][y][5]),
                'woodLvl': int(data[x][y][6]),
                'cityCount': int(data[x][y][7]),
                'piracyInRange': bool(data[x][y][8]),
                'heliosTower': bool(int(data[x][y][9])),
                'red': bool(int(data[x][y][10])),
                'blue': bool(int(data[x][y][11])),
                'resourceName': materials_names_english[int(data[x][y][2])],
                'miracleName': miracle_names_english[int(data[x][y][3])]
            })

    return worldMapIslands
# [
# "58",         //id 0
# "Phytios",    //name 1
# "1",          //resource type 2
# "2",          //type of miracle 3
# "5",          // ?? 4
# "4",          // ?? 5
# "9",          // lumber level  6
# "12",         // number of people 7
# 0,            // piracy in range 8
# "0",          // helios tower 9
# "0",          // red 10
# "0"           // blue 11
# ]

def getIsland(html: str) -> IslandDict:
    """This function uses the html passed to it as a string to extract, parse and return an Island object
    Parameters
    ----------
    html : str
        the html returned when a get request to view the island is made. This request can be made with the following statement: ``s.get(urlIsla + islandId)``, where ``urlIsla`` is a string defined in ``config.py`` and ``islandId`` is the id of the island.

    Returns
    -------
    island : Island
        this function returns a json parsed Island object.
    """
    isla = re.search(r'ajax.Responder, (\[\[[\S\s]*?\]\])\)\;', html).group(1)

    island: IslandDict = json.loads(isla)[1][1]

    # Must add aliases for different properties to maintain backwards compatibility with old code

    island["x"] = int(island["xCoord"])
    island["y"] = int(island["yCoord"])
    island["tipo"] = str(island["tradegood"])

    for city in island["cities"]:
        for key in ["Id", "Name", "AllyId", "AllyTag"]:
            if ('owner'+key) in city:
                city[key] = city['owner'+key]
        
        if "buildplace_type" in city:
            city["_type"] = city["buildplace_type"]
        
        if city["type"] == 'buildplace':
            city["type"] = 'empty'

    return island



    # isla = isla.replace("buildplace", "empty")
    # isla = isla.replace("xCoord", "x")
    # isla = isla.replace("yCoord", "y")
    # isla = isla.replace(',"owner', ',"')

    # # {"id":idIsla,"name":nombreIsla,"x":,"y":,"good":numeroBien,"woodLv":,"goodLv":,"wonder":numeroWonder, "wonderName": "nombreDelMilagro","wonderLv":"5","cities":[{"type":"city","name":cityName,"id":cityId,"level":lvIntendencia,"Id":playerId,"Name":playerName,"AllyId":,"AllyTag":,"state":"vacation"},...}}
    # isla = json.loads(isla, strict=False)
    # isla["tipo"] = re.search(r'"tradegood":(\d)', html).group(1)
    # isla["x"] = int(isla["x"])
    # isla["y"] = int(isla["y"])

    # return isla


def getCity(html: str) -> FullCityDict:
    """This function uses the ``html`` passed to it as a string to extract, parse and return a City object
    Parameters
    ----------
    html : str
        the html returned when a get request to view the city is made. This request can be made with the following statement: ``s.get(urlCiudad + id)``, where urlCiudad is a string defined in ``config.py`` and id is the id of the city.

    Returns
    -------
    city : dict
        this function returns a json parsed City object. For more information about this object refer to the github wiki page of Ikabot.
    """

    city = re.search(
        r'"updateBackgroundData",\s?([\s\S]*?)\],\["updateTemplateData"', html
    ).group(1)
    city = json.loads(city, strict=False)

    city["ownerId"] = city.pop("ownerId")
    city["ownerName"] = decodeUnicodeEscape(city["ownerName"])
    city["x"] = int(city["islandXCoord"])
    city["y"] = int(city["islandYCoord"])
    city["cityName"] = decodeUnicodeEscape(city["name"])
    city["name"] = decodeUnicodeEscape(city["name"])

    i = 0
    for position in city["position"]:
        position["position"] = i
        i += 1
        if "level" in position:
            position["level"] = int(position["level"])
        position["isBusy"] = False
        if "constructionSite" in position["building"]:
            position["isBusy"] = True
            position["building"] = position["building"][:-17]
        elif "buildingGround " in position["building"]:
            position["name"] = "empty"
            position["type"] = position["building"].split(" ")[-1]
            position["building"] = "empty"

    city["id"] = str(city["id"])
    city["isOwnCity"] = True
    city["availableResources"] = getAvailableResources(html, num=True)
    city["storageCapacity"] = getWarehouseCapacity(html)
    city["freeCitizens"] = getFreeCitizens(html)
    city["wineConsumptionPerHour"] = getWineConsumptionPerHour(html)
    city["resourcesListedForSale"] = getResourcesListedForSale(html)
    city["freeSpaceForResources"] = []
    for i in range(5):
        city["freeSpaceForResources"].append(
            city["storageCapacity"]
            - city["availableResources"][i]
            - city["resourcesListedForSale"][i]
        )

    return city

def getTransportLoadingAndTravelTime(html: str, totalResources = 0, useFreighters = False, capacityPerTransportPercent = 100, tritonBoostPercent = 0) -> tuple[int, int, int, int]:
    """Gets total loading and travel time for a shipment.
    Parameters
    ----------
    html : str
        text of the response obtained when requesting the `transport` view from the trading port
    totalResources : int
        total amount of resources that are being sent
    useFreighters : bool
        whether or not freighters will be used for this shipment (increases travel time by 20x)
    capacityPerTransportPercent : int
        percentage of the total capacity for each transport. Lowering this decreases travel time but also decreases the number of resources that can be sent. 
        Possible values are 100, 80, 60, 40, 20. These correspond to a 0, 16.7, 33.3, 50, and 66.7 % speed boost respectively
    tritonBoostPercent : int
        percentage speed boost gained from triton engines. Possible values are 0, 100, 200 and 300

    Returns
    -------
    totalTime : int
        loading time + travel time + queueTime
    loadingTime : int
        seconds it takes to load resources. This only depends on the number of resources and trading port level.
    travelTime : int
        seconds it takes to travel to the destination. Depends on distance, transporter type and world, government, triton, poseidon, sea chart archive bonuses
    queueTime : int
        seconds it takes for port to load OTHER shipments. 0 if port is not busy currently
    """
    assert capacityPerTransportPercent in [100, 80, 60, 40, 20], 'Please enter valid capacityPerTransportPercent, available values are 100, 80, 60, 40, 20'
    assert tritonBoostPercent in [0, 100, 200, 300], 'Please enter valid tritonBoostPercent, available values are 0, 100, 200, 300'
    
    # get relevant bonuses and parmeters from response
    transporterSpeed = float(re.search(r"'transporterSpeed': ([\d\.]+),", html).group(1))
    worldBonus = float(re.search(r"'worldBonus': ([\d\.]+),", html).group(1))
    governmentBonus = float(re.search(r"'governmentBonus': ([\d\.]+),", html).group(1))
    poseidonEffect = float(re.search(r"'poseidonEffect': ([\d\.]+),", html).group(1))
    marineChartArchiveBonus = float(re.search(r"'marineChartArchiveBonus': ([\d\.]+),", html).group(1))
    minimumJourneyDuration = int(re.search(r"'minimumJourneyDuration': (\d+),", html).group(1))
    distance = float(re.search(r"'distance': ([\d\.]+),", html).group(1))
    fleetJourneyTime = int(re.search(r"'fleetJourneyTime': (\d+),", html).group(1))
    queueTime = int(re.search(r"'queueTime': (\d+),", html).group(1))
    loadingSpeed = float(re.search(r"'loadingSpeed': ([\d\.]+),", html).group(1))
    
    # make sure queue time is not in the past
    queueTime = 0 if queueTime - time.time() <= 0 else int(queueTime - time.time())
    
    # calculate loading time
    loadingTime = int(totalResources / loadingSpeed)

    # calculate travel time            # lower capacity actually speeds up the transporter speed instead of lowering total travel time, this is stupid
    fleetSpeed = floor(transporterSpeed *  (1.0 + (-0.835 * capacityPerTransportPercent + 83.5) / 100 ) ) * worldBonus * governmentBonus * (1.0 + poseidonEffect + tritonBoostPercent / 100)
    uncappedDuration = int(ceil(((distance * fleetJourneyTime) / fleetSpeed) * marineChartArchiveBonus))
    uncappedDuration *= 20 if useFreighters else 1

    travelTime = uncappedDuration if uncappedDuration > minimumJourneyDuration else minimumJourneyDuration

    return  travelTime + loadingTime + queueTime, loadingTime, travelTime, queueTime



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import sys

from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import enter, read


def logs(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        while True:
            banner()
            print("0) Back")
            print("1) Set log level")
            print("2) View logs")
            choice = read(min=0, max=2, digit=True)
            if choice == 0:
                event.set()
                return
            elif choice == 1:
                # TODO changing the log level doesn't work because of multiprocessing. When the switch to multithreading is made, this will work.
                banner()
                print(
                    "The current log level is: "
                    + logging.getLevelName(logging.getLogger().getEffectiveLevel())
                    + "\n"
                )
                print("0) DEBUG")
                print("1) INFO")
                print("2) WARN")
                print("3) ERROR")
                choice = read(min=0, max=3, digit=True)
                if choice == 0:
                    logging.getLogger().setLevel(logging.DEBUG)
                elif choice == 1:
                    logging.getLogger().setLevel(logging.INFO)
                elif choice == 2:
                    logging.getLogger().setLevel(logging.WARN)
                elif choice == 3:
                    logging.getLogger().setLevel(logging.ERROR)
                print(
                    "The log level for this session has been set to: "
                    + logging.getLevelName(logging.getLogger().getEffectiveLevel())
                    + "\n"
                )
                enter()
            else:
                viewLogs()

    except KeyboardInterrupt:
        event.set()
        return


def viewLogs():
    banner()
    
    # Display last 10kb of log form logfile
    with open(LOGS_DIRECTORY_FILE, "rb") as f:
        f.seek(0, 2)
        size = f.tell()
        f.seek(max(0, size - 10000))
        print(f.read().decode("utf-8"))
    
    print(bcolors.DARK_GREEN + f"Above are displayed the last 10kb of the log file. The log file can be found at {LOGS_DIRECTORY_FILE}" + bcolors.ENDC)
    enter()


"""
This module will set up proper ikabot logging when imported.
"""

import logging
import logging.handlers

from ikabot.config import LOGS_DIRECTORY_FILE, DEFAULT_LOG_LEVEL

# TODO wrap logging functions to remove cookies from logs, or add a filter
class IkabotLogger(logging.Logger):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
logging.setLoggerClass(IkabotLogger)    

# Create custom file logger
rotatingFileHandler = logging.handlers.RotatingFileHandler(
                filename=LOGS_DIRECTORY_FILE,
                maxBytes=10 * 1024 * 1024, #max logfile size is 10 MB
                backupCount=10, # max number of log files is 10
                    )
logConfig = {
    'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    'level': DEFAULT_LOG_LEVEL,
    'force': True,
    'handlers': [rotatingFileHandler]
    }
logging.basicConfig(**logConfig)

# Make sure to set all loggers to propagate and clear their handlers (only use the root logger)
for name in logging.root.manager.loggerDict:
    logging.getLogger(name).propagate = True
    logging.getLogger(name).handlers.clear()
    
def getLogger(name: str) -> IkabotLogger:
    """Convenience function to get a logger by name"""
    return logging.getLogger(name)


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import traceback

from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.resources import getAvailableResources, getProductionPerHour
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.varios import wait, getDateTime


def donationBot(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        (cities_ids, cities) = getIdsOfCities(session)
        cities_dict = {}
        initials = [material_name[0] for material_name in materials_names]
        print(
            "Enter how often you want to donate in minutes. (min = 1, default = 1 day)"
        )
        waiting_time = read(min=1, digit=True, default=1 * 24 * 60)
        print(
            "Enter a maximum additional random waiting time between donations in minutes. (min = 0, default = 1 hour)"
        )
        max_random_waiting_time = read(min=0, digit=True, default=1 * 60)
        print(
            """Which donation method would you like to use to donate automatically? (default = 1)
(1) Donate exceeding percentage of your storage capacity
(2) Donate a percentage of production
(3) Donate specific amount
        """
        )
        donate_method = read(min=1, max=3, digit=True, default=1)
        for cityId in cities_ids:
            tradegood = cities[cityId]["tradegood"]
            initial = initials[int(tradegood)]
            print(
                
                "In {} ({}), Do you wish to donate to the forest, to the trading good, to both or none? [f/t/b/n]".format(cities[cityId]["name"], initial)
            )
            f = "f"
            t = "t"
            b = "b"
            n = "n"

            rta = read(values=[f, f.upper(), t, t.upper(), b, b.upper(), n, n.upper()])
            if rta.lower() == f:
                donation_type = "resource"
            elif rta.lower() == t:
                donation_type = "tradegood"
            elif rta.lower() == b:
                donation_type = "both"
            else:
                donation_type = None
                percentage = None

            if donation_type is not None and donate_method == 1:
                print(
                    
                    "What is the maximum percentage of your storage capacity that you wish to keep occupied? (the resources that exceed it, will be donated) (default: 80%)"
                    
                )
                percentage = read(min=0, max=100, empty=True)
                if percentage == "":
                    percentage = 80
                elif (
                    percentage == 100
                ):  # if the user is ok with the storage beeing totally full, don't donate at all
                    donation_type = None
            elif donation_type is not None and donate_method == 2:
                print(
                    
                    "What is the percentage of your production that you wish to donate? (enter 0 to disable donation for the town) (default: 50%)"
                    
                )
                percentage = read(
                    min=0, max=100, empty=True
                )  # max_random_waiting_time increases inaccuracy
                if percentage == "":
                    percentage = 50
                elif percentage == 0:
                    donation_type = None
            elif donation_type is not None and donate_method == 3:
                print(
                    
                    "What is the amount would you like to donate? (enter 0 to disable donation for the town) (default: 10000)"
                    
                )
                percentage = read(
                    min=0, empty=True
                )  # no point changing the variable's name everywhere just for this
                if percentage == "":
                    percentage = 10000
                elif percentage == 0:
                    donation_type = None

            cities_dict[cityId] = {
                "donation_type": donation_type,
                "percentage": percentage,
            }

        print("I will donate every {} minutes.".format(waiting_time))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = "\nI donate every {} minutes\n".format(waiting_time)
    setInfoSignal(session, info)
    try:
        do_it(
            session,
            cities_ids,
            cities_dict,
            waiting_time,
            max_random_waiting_time,
            donate_method,
        )
    except Exception as e:
        msg = "Error in:\n{}\nCause:\n{}".format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def do_it(
    session,
    cities_ids,
    cities_dict,
    waiting_time,
    max_random_waiting_time,
    donate_method,
):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    cities_ids : list[int]
    cities_dict : dict[int, dict]
    waiting_time: int
    max_random_waiting_time: int
    """
    for cityId in cities_ids:
        html = session.get(city_url + cityId)
        city = getCity(html)
        cities_dict[cityId]["island"] = city["islandId"]

    while True:
        total_donated = 0
        for cityId in cities_ids:
            donation_type = cities_dict[cityId]["donation_type"]
            if donation_type is None:
                continue

            # get the storageCapacity and the wood this city has
            html = session.get(city_url + cityId)
            city = getCity(html)
            wood = city["availableResources"][0]
            storageCapacity = city["storageCapacity"]

            # get the percentage
            if donate_method == 1:
                percentage = cities_dict[cityId]["percentage"]
                percentage /= 100

                # calculate what is the amount of wood that should be preserved
                max_wood = storageCapacity * percentage
                max_wood = int(max_wood)

                # calculate the wood that is exceeding the percentage
                to_donate = wood - max_wood
                if to_donate <= 0:
                    continue

            elif donate_method == 2:
                # get current production rate if changed since starting the bot
                (wood_prod, good_prod, typeGood) = getProductionPerHour(
                    session, cityId
                )
                percentage = cities_dict[cityId]["percentage"]

                # calculate the amount of wood to be donated from production, based on the given donation frequency
                to_donate = int((float(wood_prod) * percentage / 100) * (waiting_time / 60))
                # Note: Connection delay can/will cause "inaccurate" donations especially with low waiting_time
                if to_donate <= 0:
                    continue

            elif donate_method == 3:
                percentage = cities_dict[cityId]["percentage"]
                # make sure the donation amount is never lower than resources available
                max_wood = wood - percentage
                max_wood = int(max_wood)

                to_donate = percentage
                if max_wood <= 0:
                    continue

            islandId = cities_dict[cityId]["island"]
            
            # donate
            if donation_type == "both":
                forrest = int(to_donate / 2)
                trade = int(to_donate / 2)
                session.post(
                    params={
                        "islandId": islandId,
                        "type": "resource",
                        "action": "IslandScreen",
                        "function": "donate",
                        "donation": forrest,
                        "backgroundView": "island",
                        "templateView": donation_type,
                        "actionRequest": actionRequest,
                        "ajax": "1",
                    }
                )
                wait(1, maxrandom=5)  # just to simulate user interaction
                session.post(
                    params={
                        "islandId": islandId,
                        "type": "tradegood",
                        "action": "IslandScreen",
                        "function": "donate",
                        "donation": trade,
                        "backgroundView": "island",
                        "templateView": donation_type,
                        "actionRequest": actionRequest,
                        "ajax": "1",
                    }
                )
            else:
                session.post(
                    params={
                        "islandId": islandId,
                        "type": donation_type,
                        "action": "IslandScreen",
                        "function": "donate",
                        "donation": to_donate,
                        "backgroundView": "island",
                        "templateView": donation_type,
                        "actionRequest": actionRequest,
                        "ajax": "1",
                    }
                )
            total_donated += to_donate
        session.setStatus(
            f"Donated {total_donated} wood @{getDateTime()}"
        )
        msg = "I donated automatically."
        sendToBotDebug(session, msg, debugON_donationBot)

        # sleep a day
        wait(waiting_time * 60, maxrandom=max_random_waiting_time * 60)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

from ikabot.config import *
from ikabot.helpers.pedirInfo import *

# import json
# import re
# from ikabot.helpers.getJson import *
# from ikabot.helpers.gui import *
# from ikabot.helpers.resources import *
# from ikabot.helpers.varios import *


def modifyProduction(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        mod_msg = "In which cities do you want to modify production?"
        city_ids, _ = ignoreCities(session, msg=mod_msg)
        
        cities_to_process = {}
        #print("DEBUG: Fetching detailed information for selected cities...")
        for city_id in city_ids:
            html = session.get(city_url + city_id)
            city = getCity(html)
            if city['islandId'] not in cities_to_process:
                cities_to_process[city['islandId']] = {'island': getIsland(session.get(island_url + city['islandId'])), 'cities': []}
            cities_to_process[city['islandId']]['cities'].append(city)
            #print(f"DEBUG: Got city data for {city_id}")
        

        print("You want to modify the number of workers in Wood (1), Tradegood (2) or Both (3)?")
        res_choice = read(min=1, max=3)
        resource_types_to_modify = []
        if res_choice == 1:
            resource_types_to_modify.append("resource")
        elif res_choice == 2:
            resource_types_to_modify.append("tradegood")
        elif res_choice == 3:
            resource_types_to_modify.extend(["resource", "tradegood"])
            
            
        print("What % of production do you want to use? (Default 100%)")
        percentageWorkers = read(min=0, max=100, default=100)
        if percentageWorkers == 100:
            print("Would you like to use overcharge? (y/N)")
            overcharge = read(values=["y", "Y", "n", "N", ""])
            if overcharge.lower() == "y":
                use_overcharge = True
            else:
                use_overcharge = False
        else:
            use_overcharge = False
            
        banner()
        # loop through each selected city
        current_city_id_for_request = getCity(session.get())['id']
        # Loop through islands first
        for islandId, island_data in cities_to_process.items():
            island = island_data['island']
            resource_name = tradegoods_names[0]
            tradegood_name = tradegoods_names[int(island["tipo"])]
            #print(f"DEBUG: Visited island {islandId}")

            # Then loop through cities on that island
            for city in island_data['cities']:
                # Change the session context to the current city
                session.post(params={
                    'action': 'header', 'function': 'changeCurrentCity',
                    'actionRequest': actionRequest, 'cityId': city['id'],
                    'oldView': 'city', 'backgroundView': 'city',
                    'currentCityId': current_city_id_for_request, 'ajax': '1'
                })
                # Update the current city ID for the next request
                current_city_id_for_request = city['id']
                #print(f"DEBUG: session.post was made, current city is {city['id']} and island is {islandId}")

                # Loop through only the resource types the user chose to modify
                for resource_type in resource_types_to_modify:
                    # Fetch data, calculate, and set workers in one sequence
                    
                    # Fetch data for the current resource type
                    url = f"view={resource_type}&type={resource_type}&islandId={islandId}&cityId={city['id']}&backgroundView=island&currentIslandId={islandId}&actionRequest={actionRequest}&ajax=1"
                    resp = session.post(url)
                    resp_json = json.loads(resp, strict=False)
                    
                    template_data = resp_json[2][1]
                    slider_data = template_data['js_ResourceSlider']['slider']
                    max_normal = slider_data['max_value']
                    total_max = slider_data['max_value'] + slider_data['overcharge']
                    #print(f"DEBUG: session.post was made, current city is {city['id']}, current island is {islandId}\nresource type is {resource_type}, max_normal is {max_normal} and total_max is {total_max}")
                    
                    # Calculate finalWorkers using the data we just fetched
                    if percentageWorkers == 100:
                        finalWorkers = total_max if use_overcharge else max_normal
                    else:
                        finalWorkers = int(max_normal / 100 * percentageWorkers)
                    
                    # Set the workers
                    worker_key = "rw" if resource_type == "resource" else "tw"    
                    session.post(params={
                        "islandId": islandId, "cityId": city["id"],
                        "type": resource_type, "screen": resource_type,
                        "action": "IslandScreen", "function": "workerPlan",
                         worker_key: finalWorkers, "templateView": resource_type,
                        "actionRequest": actionRequest, "ajax": "1"
                    })
                    
                    selected_good_name = resource_name if resource_type == "resource" else tradegood_name
                    print(f"{finalWorkers} workers set for {selected_good_name} in {city['name']}.")
                
        print("\nAll productions have been set!")
        enter()
        event.set()
        
    except KeyboardInterrupt:
        event.set()
        return 



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import os
import re
import sys
from decimal import *

from ikabot import config
from ikabot.config import *
from ikabot.helpers.getJson import *
from ikabot.helpers.gui import *
from ikabot.helpers.varios import decodeUnicodeEscape

getcontext().prec = 30


def read(
    min=None,
    max=None,
    digit=False,
    msg=prompt,
    values=None,
    empty=False,
    additionalValues=None,
    default=None,
    _retries=0,
    _max_retries=20
):
    """Reads input from user
    Parameters
    ----------
    min : int
        smallest number acceptable as input
    max : int
        greatest number acceptable as input
    digit : bool
        boolean indicating whether or not the input MUST be an int
    msg : str
        string printed before the user is asked for input
    values : list
        list of strings which are acceptable as input
    empty : bool
        a boolean indicating whether or not an empty string is acceptable as input
    additionalValues : list
        list of strings which are additional valid inputs. Can be used with digit = True to validate a string as an input among all digits
    Returns
    -------
    result : int | str
        int representing the user's choice
    """
    if _retries >= _max_retries:
        print('Error: Max recursion depth reached for read function. Returning None!')
        return None
        
    if min is not None and max is not None and min > max:
        print('Error: minimum input value is greater than maximum input value!')
        return None
        
    try:
        if len(config.predetermined_input) != 0:
            return config.predetermined_input.pop(0)
    except Exception:
        pass
    
    def _invalid():
        print("\033[1A\033[K", end="")  # remove line
        return read(min=min, max=max, digit=digit, msg=msg, values=values, empty=empty, additionalValues=additionalValues, default=default, _retries=_retries+1, _max_retries=_max_retries)
    
    try:
        read_input = input(msg)
    except EOFError:
        return _invalid()
    
    if additionalValues is not None and read_input in additionalValues:
        return read_input
    if read_input == "" and default is not None:
        return default
    if read_input == "" and empty is True:
        return read_input
    
    if digit is True or min is not None or max is not None:
        if not read_input.isdigit():
            return _invalid()
        else:
            try:
                read_input = int(read_input)
            except ValueError:
                return _invalid()
    
    if min is not None and read_input < min:
        return _invalid()
    if max is not None and read_input > max:
        return _invalid()
    if values is not None and read_input not in values:
        return _invalid()
    return read_input


def chooseCity(session, foreign=False):
    """Prompts the user to chose a city
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    foreign : bool
        lets the user choose a foreign city

    Returns
    -------
    city : City
        a city object representing the chosen city
    """
    global menu_cities
    (ids, cities) = getIdsOfCities(session)
    if menu_cities == "":
        longest_city_name_length = 0
        for city_id in ids:
            length = len(cities[city_id]["name"])
            if length > longest_city_name_length:
                longest_city_name_length = length

        def pad(city_name):
            return " " * (longest_city_name_length - len(city_name) + 2)

        resources_abbreviations = {
            "1": "(W)",
            "2": "(M)",
            "3": "(C)",
            "4": "(S)",
        }

        i = 0
        if foreign:
            print(" 0: foreign city")
        else:
            print("")
        for city_id in ids:
            i += 1
            resource_index = str(cities[city_id]["tradegood"])
            resource_abb = resources_abbreviations[resource_index]
            city_name = decodeUnicodeEscape(cities[city_id]["name"])
            menu_cities += "{: >2}: {}{}{}\n".format(
                i, city_name, pad(city_name), resource_abb
            )
        menu_cities = menu_cities[:-1]
    if foreign:
        print(" 0: foreign city")
    print(menu_cities)

    if foreign:
        selected_city_index = read(min=0, max=len(ids))
    else:
        selected_city_index = read(min=1, max=len(ids))
    if selected_city_index == 0:
        return chooseForeignCity(session)
    else:
        html = session.get(city_url + ids[selected_city_index - 1])
        return getCity(html)


def chooseForeignCity(session):
    """Prompts the user to select an island, and a city on that island (is only used in chooseCity)
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    city : City
        a city object representing the city the user chose
    """
    banner()
    x = read(msg="coordinate x:", digit=True)
    y = read(msg="coordinate y:", digit=True)
    print("")
    url = "view=worldmap_iso&islandX={}&islandY={}&oldBackgroundView=island&islandWorldviewScale=1".format(
        x, y
    )
    html = session.get(url)
    try:
        islands_json = re.search(r"jsonData = \'(.*?)\';", html).group(1)
        islands_json = json.loads(islands_json, strict=False)
        island_id = islands_json["data"][str(x)][str(y)][0]
    except Exception:
        print("Incorrect coordinates")
        enter()
        banner()
        return chooseCity(session, foreign=True)
    html = session.get(island_url + island_id)
    island = getIsland(html)

    i = 0
    city_options = []
    for city in island["cities"]:
        if (
            city["type"] == "city"
            and city["state"] == ""
            and city["Name"] != session.username
        ):
            i += 1
            num = " " + str(i) if i < 10 else str(i)
            print(
                "{: >2}: {: >{max_city_name_length}} ({})".format(
                    num,
                    decodeUnicodeEscape(city["name"]),
                    decodeUnicodeEscape(city["Name"]),
                    max_city_name_length=MAXIMUM_CITY_NAME_LENGTH,
                )
            )
            city_options.append(city)
    if i == 0:
        print("There are no cities where to send resources on this island")
        enter()
        return chooseCity(session, foreign=True)
    selected_city_index = read(min=1, max=i)
    city = city_options[selected_city_index - 1]
    city["islandId"] = island["id"]
    city["cityName"] = decodeUnicodeEscape(city["name"])
    city["isOwnCity"] = False
    return city


def askForValue(text, max_val):
    """Displays text and asks the user to enter a value between 0 and max

    Parameters
    ----------
    text : str
        text to be displayed when asking the user for input
    max_val : int
        integer representing the number of input options

    Returns
    -------
    var : int
        integer representing the user's input
        if the user has inputed nothing, 0 will be returned instead
    """
    var = read(
        msg=text, min=0, max=max_val, default=0, additionalValues=["all", "half"]
    )
    if var == "all":
        var = max_val
    elif var == "half":
        var = max_val // 2
    return var


def getIdsOfCities(session, all=False):
    """Gets the user's cities
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    all : bool
        boolean indicating whether all cities should be returned, or only those that belong to the current user

    Returns
    -------
    (ids, cities) : tuple
        a tuple containing the a list of city IDs and a list of city objects
    """
    global cities_cache
    global ids_cache
    if ids_cache is None or cities_cache is None or session.padre is False:
        html = session.get()
        cities_cache = (
            re.search(
                r'relatedCityData:\sJSON\.parse\(\'(.+?),\\"additionalInfo', html
            ).group(1)
            + "}"
        )
        cities_cache = cities_cache.replace("\\", "")
        cities_cache = cities_cache.replace("city_", "")
        cities_cache = json.loads(cities_cache, strict=False)

        ids_cache = [city_id for city_id in cities_cache]
        ids_cache = sorted(ids_cache)

    # {'coords': '[x:y] ', 'id': idCiudad, 'tradegood': '..', 'name': 'nomberCiudad', 'relationship': 'ownCity'|'occupiedCities'|..}
    if all is False:
        ids_own = [
            city_id
            for city_id in cities_cache
            if cities_cache[city_id]["relationship"] == "ownCity"
        ]
        ids_other = [
            city_id
            for city_id in cities_cache
            if cities_cache[city_id]["relationship"] != "ownCity"
        ]
        own_cities = cities_cache.copy()
        for id in ids_other:
            del own_cities[id]
        return ids_own, own_cities
    else:
        return ids_cache, cities_cache


def getIslandsIds(session):
    """Gets the IDs of islands the user has cities on
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    islands_ids : list
        a list containing the IDs of the users islands
    """
    (cities_ids, cities) = getIdsOfCities(session)
    islands_ids = set()
    for city_id in cities_ids:
        html = session.get(city_url + city_id)
        city = getCity(html)
        island_id = city["islandId"]
        islands_ids.add(island_id)
    return list(islands_ids)


def getShipCapacity(session):
    """Gets the ship capacity
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    capacity : int
        an integer representing the ship capacity of the user's current city
    """
    html = session.post('view=merchantNavy')
    ship_capacity = html.split('singleTransporterCapacity":')[1].split(',"singleFreighterCapacity')[0]
    freighter_capacity = html.split('singleFreighterCapacity":')[1].split(',"draftEffect')[0]

    return int(ship_capacity), int(freighter_capacity)


def ignoreCities(session, msg=None):
    """Prompts the user to select cities which should be ignored
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    (cities_ids) : list
        a list containing IDs of cities that were not ignored
    """
    (cities_ids, cities) = getIdsOfCities(session)
    choice = None
    ignored_cities = []
    while True:
        banner()
        if msg is not None: print(f"{msg}")
        displayed_string = (
            f'(currently ignoring: {", ".join(ignored_cities)})'
            if ignored_cities
            else ""
        )
        print(f"Select cities to ignore. {displayed_string}")
        print("0) Continue")
        choice_to_cityid_map = []
        for i, city in enumerate(cities.values()):
            choice_to_cityid_map.append(city["id"])
            print(f'{i + 1}) {city["name"]} - {materials_names[city["tradegood"]]}')
        choice = read(min=0, max=len(cities_ids))
        if choice == 0:
            break
        city_id = choice_to_cityid_map[choice - 1]
        cities_ids = list(filter(lambda x: x != str(city_id), cities_ids))
        ignored_cities.append(cities[str(city_id)]["name"])
        del cities[str(city_id)]
        
    return cities_ids, cities



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import sys

from ikabot.helpers.botComm import sendToBot
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import enter, read


def testTelegramBot(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        input = read(msg="Enter the massage you wish to see: ")
        msg = "Test message: {}".format(input)
        sendToBot(session, msg)
        enter()
        event.set()
    except KeyboardInterrupt:
        event.set()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import re
import time

from ikabot.config import *
from ikabot.helpers.getJson import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.resources import *
from ikabot.helpers.varios import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.botComm import *

wait_time = 60 * 60 * 12  # 12 hours, wait_time*6 equals total shrine grace time of 72h
last_donation_status = ""
last_donation_time = ""
current_favor = ""
selected_gods = ""


def findShrine(session):
    """Finding the city with Shrine and it's position
    Parameters
    ----------
    session : ikabot.web.session.Session

    Returns
    -------
    shrineCity[int], shrinePos[int]
    """
    cities_ids = getIdsOfCities(session)[0]
    for city_id in cities_ids:
        html = session.get(city_url + city_id)
        city = getCity(html)
        for pos, building in enumerate(city["position"]):
            if building["building"] == "shrineOfOlympus":
                shrineCity = city_id
                shrinePos = pos
                return (
                    shrineCity,
                    shrinePos,
                )  # Return as soon as Shrine and it's position is found
    return None, None  # Return None if not found


def do_it(session, godids, mode, times):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    godids : list[int]
    mode : int
    times : int
    """
    try:
        favor_needed = (
            len(godids) * 100
        )  # Calculate the required amount of favor to donate all selected gods at once
        shrineCity, shrinePos = findShrine(session)
        if shrineCity is not None and shrinePos is not None:
            cityid, pos = shrineCity, shrinePos
        else:
            msg = "Shrine city or building position was not found."
            sendToBot(session, msg)
            event.set()
            return

        if mode == 1 or mode == 3:
            for _ in range(times):
                favor = getFavor(session, cityid, pos)
                while favor < favor_needed:
                    session.setStatus(
                        f"Not enough favor @{getDateTime()}, re-trying in 3h."
                    )
                    time.sleep(wait_time / 4)  # 12h / 4 = 3 hours
                    favor = getFavor(session, cityid, pos)
                for godid in godids:
                    donateShrine(session, godid, cityid, pos)
                    time.sleep(2)

            if mode == 1:
                event.set()
                return
            mode = 2  # If mode is both, set mode for loop

        if mode == 2:
            while True:
                favor = getFavor(session, cityid, pos)
                while favor < favor_needed:
                    session.setStatus(
                        f"Not enough favor @{getDateTime()}, re-trying in 3h."
                    )
                    time.sleep(wait_time / 4)  # 12h / 4 = 3 hours
                    favor = getFavor(session, cityid, pos)
                for godid in godids:
                    donateShrine(session, godid, cityid, pos)
                    time.sleep(2)
                for i in range(6):  # 12h * 6 = 72 hours
                    time.sleep(wait_time)  # 12h
                    global current_favor
                    current_favor = getFavor(session, cityid, pos)
                    global last_donation_status
                    last_donation_status = f"Activated {selected_gods} @{last_donation_time}, current favor: {current_favor}"
                    session.setStatus(
                        last_donation_status
                    )  # Update only current favor value in the task status message, activation/donation time remains

    except Exception as e:
        msg = f"Error in activateShrine:\n\n{e}"
        sendToBot(session, msg)
        event.set()
        return


def activateShrine(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd : int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    banner()
    godids = []

    while True:
        print(
            """Which God(s) would you like to activate autonomously? 
    (0) Continue
    (1) Pan (Wood)
    (2) Dionysus (Wine)
    (3) Tyche (Marble)
    (4) Plutus (Gold)
    (5) Theia (Crystal)
    (6) Hephaestus (Sulphur)
    """
        )
        god = read(min=0, max=6, digit=True)
        if god == 0:
            if len(godids) == 0:
                event.set()
                return
            named_gods = [gods(godid) for godid in godids]
            gods_str = ", ".join(named_gods) if len(named_gods) > 1 else named_gods[0]
            global selected_gods
            selected_gods = gods_str
            break
        else:
            godids.append(god)
            continue

    print("")
    print(
        """Would you like to activate the selected God(s) a specific amount of times or autonomously every 70 hours?
(1) Specific amount of times in a row
(2) Autonomously every 70 hours
(3) Both
"""
    )
    mode = read(min=1, max=3, digit=True)
    if mode == 1 or mode == 3:
        print("How many times would you like to activate the selected God(s)?")
        times = read(min=1, max=10, digit=True)
    if mode == 2:
        mode = 2
        times = 0

    set_child_mode(session)
    event.set()

    try:
        do_it(
            session,
            tuple(godids),
            mode,
            times,
        )

    except Exception as e:
        msg = f"Error in activateShrine:\n\n{e}"
        sendToBot(session, msg)
    finally:
        session.logout()


def gods(godid):
    """Converts godid into it's respectable name
    Parameters
    ----------
    session : ikabot.web.session.Session
    godid : int
    """
    names = {
        1: "Pan",
        2: "Dionysus",
        3: "Tyche",
        4: "Plutus",
        5: "Theia",
        6: "Hephaestus",
    }
    return names.get(godid, None)


def getFavor(session, cityid, pos):
    """Extracts the currentFavor amount from Ikariam
    Parameters
    ----------
    session : ikabot.web.session.Session
    cityid : int
    pos : int
    Returns
    -------
    favor[int]
    """
    try:
        fav_url = f"view=shrineOfOlympus&cityId={cityid}&position={pos}&activeTab=tabOverview&backgroundView=city&currentCityId={cityid}&templateView=shrineOfOlympus&actionRequest={actionRequest}&ajax=1"
        get_fav = session.get(fav_url, noIndex=True)
        load_fav = json.loads(get_fav, strict=False)
        favor = load_fav[2][1]["currentFavor"]
        return favor
    except Exception as e:
        print(e)
        return 0


def donateShrine(session, godid, cityid, pos):
    """Donates to the selected Gods and updates the task status accordingly
    Parameters
    ----------
    session : ikabot.web.session.Session
    godid : int
    cityid : int
    pos : int
    """
    url = f"action=DonateFavorToGod&godId={godid}&position={pos}&backgroundView=city&currentCityId={cityid}&templateView=shrineOfOlympus&currentTab=tabOverview&actionRequest={actionRequest}&ajax=1"
    session.post(url)
    global current_favor
    current_favor = getFavor(session, cityid, pos)
    global last_donation_time
    last_donation_time = getDateTime()
    global last_donation_status
    last_donation_status = f"Activated {selected_gods} @{last_donation_time}, current favor: {current_favor}"
    session.setStatus(last_donation_status)



import setuptools

from ikabot.config import IKABOT_VERSION

setuptools.setup(
    name="ikabot",
    version=IKABOT_VERSION,
    author="physics-sp",
    author_email="physics-sp@protonmail.com",
    license="MIT",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Ikabot-Collective/ikabot",
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=["requests", "requests[socks]", "cryptography", "psutil"],
    extras_require={"webServer": ["flask"]},
    entry_points={
        "console_scripts": ["ikabot=ikabot.command_line:main"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import datetime
import sys

from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.pedirInfo import enter, read
from ikabot.helpers.process import run, updateProcessList


def killTasks(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        while True:
            banner()
            process_list = updateProcessList(session)
            process_list = [
                process for process in process_list if process["action"] != "killTasks"
            ]
            if len(process_list) == 0:
                print("There are no tasks running")
                enter()
                event.set()
                return
            print("Which task do you wish to kill?\n")
            print("(0) Exit")
            for process in process_list:
                if "date" in process:
                    print(
                        "({}) {:<35}{:>20} {:>10}".format(
                            process_list.index(process) + 1,
                            process["action"],
                            datetime.datetime.fromtimestamp(process["date"]).strftime(
                                "%b %d %H:%M:%S"
                            ),
                            process["pid"]
                        )
                    )
                else:
                    print(
                        "({}) {:<35} {:>10}".format(
                            process_list.index(process) + 1,
                            process["action"],
                            process["pid"]
                        )
                    )
            choise = read(min=0, max=len(process_list), digit=True)
            if choise == 0:
                event.set()
                return
            else:
                if isWindows:
                    run("taskkill /F /PID {}".format(process_list[choise - 1]["pid"]))
                else:
                    run("kill -9 {}".format(process_list[choise - 1]["pid"]))
    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import traceback
from ikabot.helpers.pedirInfo import read, enter
from ikabot.helpers.gui import *
from ikabot.config import *

from importlib.machinery import SourceFileLoader

def loadCustomModule(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        sessionData = session.getSessionData()
        modules = [path for path in sessionData.get('shared', {}).get('customModules', []) if os.path.isfile(path)]
        print("0) Back")
        print("1) Add new module")
        for module in modules:
            print(str(modules.index(module) + 2) + ") " + module)

        choice = read(min = 0, max = len(modules) + 1, digit = True)
        if choice == 0:
            event.set()
            return
        elif choice == 1:
            banner()
            print(f'        {bcolors.WARNING}[WARNING]{bcolors.ENDC} Running third party code could be dangerous!\n\n')
            print('Enter the full path to the module you wish to load. The module must have a function with the same name as the file.')
            print('Enter full path: ')
            path = read().strip().replace('\\', '/')
            if not path.endswith('.py'):
                print('The file must be a .py file!')
                enter()
                event.set()
                return

            modules.append(path)
            sessionData = session.getSessionData()
            sessionData['shared'] = sessionData.get('shared', {})
            sessionData['shared']['customModules'] = modules
            session.setSessionData(sessionData['shared'], shared = True)

        else:    
            path = modules[choice - 2]

        name = path.split('/')[-1].split('.')[0]

        banner()

        # Load module
        try:
            temp = SourceFileLoader(name,path).load_module()
        except Exception as e:
            print(f'Error while loading {path}: ' + str(e) + '\n' + traceback.format_exc())
            enter()
            event.set()
            return
        
        # Call the function with the same name as the file
        try:
            exec('temp.' + name + '(session, event, stdin_fd, predetermined_input)')
        except Exception as e:
            print(f'Error while running {name} in {path}: ' + str(e) + '\n' + traceback.format_exc())
            enter()
            event.set()
            return
        
        event.set()
        return

      
    except KeyboardInterrupt:
        event.set()
        return
    


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import base64
import hashlib
import json
import os

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from ikabot.config import *
from ikabot.helpers.botComm import *


class AESCipher:

    def __init__(self, mail, password):
        if type(password) == int:
            password = str(password)
        self.key = hashlib.sha256(
            mail.encode("utf-8") + b"\x00" + password.encode("utf-8")
        ).digest()
        for i in range(0xFFF):
            self.key = hashlib.sha256(self.key).digest()

    def encrypt(self, plaintext):
        aesgcm = AESGCM(self.key)
        nonce = os.urandom(16)
        ciphertext = aesgcm.encrypt(nonce, plaintext.encode("utf-8"), None)
        return base64.b64encode(nonce + ciphertext).decode("utf-8")

    def decrypt(self, ciphertext):
        ciphertext = base64.b64decode(ciphertext)
        nonce = ciphertext[:16]
        ciphertext = ciphertext[16:]
        aesgcm = AESGCM(self.key)
        plaintext = aesgcm.decrypt(nonce, ciphertext, None)
        return plaintext.decode("utf-8")

    def getEntryKey(self, session):
        """
        Parameters
        ----------
        session : ikabot.web.session.Session

        Returns
        -------
        entry key : str
        """
        return hashlib.sha256(
            "ikabot".encode("utf-8") + session.mail.encode("utf-8")
        ).hexdigest()

    def deleteSessionData(self, session):
        """
        Parameters
        ----------
        session : ikabot.web.session.Session
        """
        entry_key = self.getEntryKey(session)
        with open(ikaFile, "r") as filehandler:
            data = filehandler.read()

        newFile = ""
        for line in data.split("\n"):
            if entry_key != line[:64]:
                newFile += line + "\n"

        with open(ikaFile, "w") as filehandler:
            filehandler.write(newFile.strip())
            filehandler.flush()

    def getSessionData(self, session, all=False):
        """
        Parameters
        ----------
        session : ikabot.web.session.Session
        all : bool
        """
        entry_key = self.getEntryKey(session)
        with open(ikaFile, "r") as filehandler:
            ciphertexts = filehandler.read()

        for ciphertext in ciphertexts.split("\n"):
            if entry_key == ciphertext[:64]:
                ciphertext = ciphertext[64:]
                try:
                    plaintext = self.decrypt(ciphertext)
                except Exception:
                    msg = "Error while decrypting session data.\nYou may have entered a wrong password."
                    
                    if session.padre:
                        print(msg)
                    else:
                        sendToBot(session, msg)
                    print("\nWould you like to delete the ikabot session data associated with this email address? [y/N]")
                    rta = read(values=["n", "N", "y", "Y"])
                    if rta.lower() == "n":
                        os._exit(0)
                    self.deleteSessionData(session)
                    os._exit(0)
                data_dict = json.loads(plaintext, strict=False)
                if all:
                    return data_dict
                else:
                    try:
                        try:
                            session_data = data_dict[session.username][session.mundo][
                                session.servidor
                            ]
                        except Exception:
                            session_data = {}
                        session_data["shared"] = data_dict["shared"]
                        return session_data
                    except KeyError:
                        return {}
        return {}

    def setSessionData(self, session, data, shared=False):
        """
        Parameters
        ----------
        session : ikabot.web.session.Session
        data : dict
        """
        session_data = self.getSessionData(session, True)

        if shared:
            if "shared" not in session_data:
                session_data["shared"] = {}
            if "logLevel" not in session_data["shared"]:
                session_data["shared"]["logLevel"] = 2  # Warn by default
            session_data["shared"] = {**session_data["shared"], **data}
        else:
            if session.username not in session_data:
                session_data[session.username] = {}
            if session.mundo not in session_data[session.username]:
                session_data[session.username][session.mundo] = {}
            if session.servidor not in session_data[session.username][session.mundo]:
                session_data[session.username][session.mundo][session.servidor] = {}
            if "shared" not in session_data:
                session_data["shared"] = {}
            session_data[session.username][session.mundo][session.servidor] = data

        plaintext = json.dumps(session_data)
        ciphertext = self.encrypt(plaintext)

        with open(ikaFile, "r") as filehandler:
            data = filehandler.read()

        entry_key = self.getEntryKey(session)
        newFile = ""
        newline = entry_key + " " + ciphertext
        for line in data.split("\n"):
            if entry_key != line[:64]:
                newFile += line + "\n"
        newFile += newline + "\n"

        with open(ikaFile, "w") as filehandler:
            filehandler.write(newFile.strip())
            filehandler.flush()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import random
import re
import time
from datetime import datetime
from decimal import *

getcontext().prec = 30


def addThousandSeparator(num, character="."):
    """Formats the number into a string and adds a `character` for every thousand (eg. 3000 -> 3.000)
    Parameters
    ----------
    num : int
        integer number to format
    character : str
        character to act as the thousand separator

    Returns
    -------
    number : str
        a string representing that number with added `character` for every thousand
    """
    return "{0:,}".format(int(num)).replace(",", character)


def daysHoursMinutes(totalSeconds):
    """Formats the total number of seconds into days hours minutes (eg. 321454 -> 3D 17H)
    Parameters
    ----------
    totalSeconds : int
        total number of seconds

    Returns
    -------
    text : str
        formatted string (D H M)
    """
    if totalSeconds == 0:
        return "0 s"
    dias = int(totalSeconds / Decimal(86400))
    totalSeconds -= dias * Decimal(86400)
    horas = int(totalSeconds / Decimal(3600))
    totalSeconds -= horas * Decimal(3600)
    minutos = int(totalSeconds / Decimal(60))
    texto = ""
    if dias > 0:
        texto = str(dias) + "D "
    if horas > 0:
        texto = texto + str(horas) + "H "
    if minutos > 0 and dias == 0:
        texto = texto + str(minutos) + "M "
    return texto[:-1]


def wait(seconds, maxrandom=0):
    """This function will wait the provided number of seconds plus a random number of seconds between 0 and maxrandom
    Parameters
    -----------
    seconds : int
        the number of seconds to wait for
    maxrandom : int
        the maximum number of additional seconds to wait for
    """
    if seconds <= 0:
        return
    randomTime = random.randint(0, maxrandom)
    ratio = (1 + 5**0.5) / 2 - 1  # 0.6180339887498949
    comienzo = time.time()
    fin = comienzo + seconds
    restantes = seconds
    while restantes > 0:
        time.sleep(restantes * ratio)
        restantes = fin - time.time()
    time.sleep(randomTime)


def getCurrentCityId(session):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    """
    html = session.get()
    return re.search(r"currentCityId:\s(\d+),", html).group(1)


def getDateTime(timestamp=None):
    """Returns a string of the current date and time in the YYYY-mm-dd_HH-MM-SS, if `timestamp` is provided then it converts it into the given format.
    Parameters
    ----------
    timestamp : int
        Unix timestamp to be converted

    Returns
    -------
    text : str
        Formatted string YYYY-mm-dd_HH-MM-SS
    """
    timestamp = timestamp if timestamp else time.time()
    return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d_%H-%M-%S")


def normalizeDicts(list_of_dicts):
    """Returns a list of dicts that all have the same keys. Keys will be initialized to None
    Parameters
    ----------
    list_of_dicts : [dict]
        List of dicts that may have different keys (one dict has some keys that another doesn't)

    Returns
    -------
    normalized_dicts : [dict]
        List of dicts that all have the same keys, with new ones initialized to None.
    """
    all_keys = set().union(*[d.keys() for d in list_of_dicts])
    return [{k: (d[k] if k in d else None) for k in all_keys} for d in list_of_dicts]


def decodeUnicodeEscape(input_string):
    """
    Replace Unicode escape sequences (e.g., u043c) with corresponding UTF-8 characters.

    Parameters:
    - input_string (str): The original string.

    Returns:
    - str: The string with replaced Unicode escape sequences.
    """
    return re.sub(
        r"u([0-9a-fA-F]{4})", lambda x: chr(int(x.group(1), 16)), input_string
    )


def timeStringToSec(time_string):
    """Returns number of seconds from a time string (eg. 5h 35m -> 20100s)
    Parameters
    ----------
    time_string : str
        String that needs to be converted to number of seconds

    Returns
    -------
    seconds : int
        Number of seconds
    """
    hours = re.search(r"(\d+)h", time_string)
    if hours is None:
        hours = 0
    else:
        hours = int(hours.group(1)) * 3600
    minutes = re.search(r"(\d+)m", time_string)
    if minutes is None:
        minutes = 0
    else:
        minutes = int(minutes.group(1)) * 60
    seconds = re.search(r"(\d+)s", time_string)
    if seconds is None:
        seconds = 0
    else:
        seconds = int(seconds.group(1)) * 1
    return hours + minutes + seconds

def lastloginTimetoString(time_string):
    """Returns formatet last Login String
    Parameters
    ----------
    time_string : str
        last Login String that needs to be converted

    Returns
    -------
    - str: formatet String of last Login 
    """
    date_format = '%Y-%m-%dT%H:%M:%S%z'
    lastlogin_string = time_string
    lastlogin_object = datetime.strptime(lastlogin_string, date_format)
    return lastlogin_object.strftime('%Y-%m-%d')



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import traceback

from requests import get, post

from ikabot.config import *
from ikabot.helpers.dns import getAddress


def getNewBlackBoxToken(session):
    """This function returns a newly generated blackbox token from the API
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object

    Returns
    -------
    token : str
        blackbox token
    """
    address = (
        getAddress(publicAPIServerDomain)
        + "/v1/token"
        + "?user_agent="
        + session.user_agent
    )
    response = get(address, verify=do_ssl_verify, timeout=900)
    assert response.status_code == 200, (
        "API response code is not OK: "
        + str(response.status_code)
        + "\n"
        + response.text
    )
    response = response.json()
    if "status" in response and response["status"] == "error":
        raise Exception(response["message"])
    return "tra:" + response


def getInteractiveCaptchaSolution(session, text_image, icons_image):
    """This function returns the solution of the interactive captcha
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    icons_image : bytes
        the image that contains 4 other images
    text_image : bytes
        the image that contaisn the text

    Returns
    -------
    solution : str
        solution of the captcha
    """
    address = getAddress(publicAPIServerDomain) + "/v1/decaptcha/lobby"
    files = {"text_image": text_image, "icons_image": icons_image}
    response = post(address, files=files, verify=do_ssl_verify, timeout=900)
    assert response.status_code == 200, (
        "API response code is not OK: "
        + str(response.status_code)
        + "\n"
        + response.text
    )
    response = response.json()
    if type(response) is not int:
        raise Exception('Response was not a digit as expected: ' + response)
#    if "status" in response and response["status"] == "error":
#        raise Exception(response["message"])
    return response


def getPiratesCaptchaSolution(session, image):
    """This function returns the solution of the pirates captcha
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    image : bytes
        the image to be solved

    Returns
    -------
    solution : str
        solution of the captcha
    """
    address = getAddress(publicAPIServerDomain) + "/v1/decaptcha/pirate"
    files = {"image": image}
    response = post(address, files=files, verify=do_ssl_verify, timeout=900)
    assert response.status_code == 200, (
        "API response code is not OK: "
        + str(response.status_code)
        + "\n"
        + response.text
    )
    response = response.json()
    if "status" in response and response["status"] == "error":
        raise Exception(response["message"])
    return response



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

from ikabot.helpers.logging import getLogger
import base64
import datetime
import getpass
import json
import os
import random
import re
import sys
import time
import traceback
from collections import deque

import requests
from urllib3.exceptions import InsecureRequestWarning

from ikabot import config
from ikabot.config import *
from ikabot.helpers.aesCipher import *
from ikabot.helpers.botComm import *
from ikabot.helpers.getJson import getCity
from ikabot.helpers.gui import banner
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.varios import getDateTime, lastloginTimetoString
from ikabot.helpers.apiComm import getInteractiveCaptchaSolution, getNewBlackBoxToken


class Session:
    def __init__(self):
        self.padre = True
        self.logged = False
        self.blackbox = None
        self.logger = getLogger(__name__)
        self.requestHistory = deque(maxlen=5)  # keep last 5 requests in history
        # disable ssl verification warning
        requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)
        self.__login()

    def setStatus(self, message):
        """This function will modify the current tasks status message that appears in the table on the main menu
        Parameters
        ----------
        message : Message to be displayed in the table in main menu
        """
        self.logger.info(f"Changing status to {message}")

        # read from file
        sessionData = self.getSessionData()
        try:
            fileList = sessionData["processList"]
        except KeyError:
            fileList = []
        # modify current process' status message
        [p.update({"status": message}) for p in fileList if p["pid"] == os.getpid()]
        # dump back to session data
        sessionData["processList"] = fileList
        self.setSessionData(sessionData)

    def __genRand(self):
        return hex(random.randint(0, 65535))[2:]

    def __genCookie(self):
        return (
            self.__genRand()
            + self.__genRand()
            + hex(int(round(time.time() * 1000)))[2:]
            + self.__genRand()
            + self.__genRand()
        )

    def __fp_eval_id(self):
        return (
            self.__genRand()
            + self.__genRand()
            + "-"
            + self.__genRand()
            + "-"
            + self.__genRand()
            + "-"
            + self.__genRand()
            + "-"
            + self.__genRand()
            + self.__genRand()
            + self.__genRand()
        )

    def __logout(self, html):
        if html is not None:
            idCiudad = getCity(html)["id"]
            token = re.search(r'actionRequest"?:\s*"(.*?)"', html).group(1)
            urlLogout = "action=logoutAvatar&function=logout&sideBarExt=undefined&startPageShown=1&detectedDevice=1&cityId={0}&backgroundView=city&currentCityId={0}&actionRequest={1}".format(
                idCiudad, token
            )
            self.s.get(self.urlBase + urlLogout, verify=config.do_ssl_verify)

    def __isInVacation(self, html):
        return "nologin_umod" in html

    def __isExpired(self, html):
        return "index.php?logout" in html or '<a class="logout"' in html

    def isExpired(self, html):
        return self.__isExpired(html)

    def __saveNewCookies(self):
        sessionData = self.getSessionData()

        cookie_dict = dict(self.s.cookies.items())
        sessionData["cookies"] = cookie_dict

        self.setSessionData(sessionData)

    def __getCookie(self, sessionData=None):
        if sessionData is None:
            sessionData = self.getSessionData()
        try:
            cookie_dict = sessionData["cookies"]
            self.s = requests.Session()
            self.__update_proxy(sessionData=sessionData)
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            requests.cookies.cookiejar_from_dict(
                cookie_dict, cookiejar=self.s.cookies, overwrite=True
            )
        except KeyError:
            self.__login(3)

    def __test_lobby_cookie(self):
        if "gf-token-production" in self.s.cookies:
            self.headers = {
                "Host": "lobby.ikariam.gameforge.com",
                "User-Agent": self.user_agent,
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate",
                "DNT": "1",
                "Connection": "close",
                "Referer": "https://lobby.ikariam.gameforge.com/",
                "Authorization": "Bearer " + self.s.cookies["gf-token-production"],
            }
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            if (
                self.s.get(
                    "https://lobby.ikariam.gameforge.com/api/users/me"
                ).status_code
                == 200
            ):
                return True
            self.s.cookies.clear()
        return False

    def __test_server_maintenace(self, html):
        match = re.search(
            r'\[\["provideFeedback",\[{"location":1,"type":11,"text":([\S\s]*)}\]\]\]',
            html,
        )
        if (
            match
            and '[["provideFeedback",[{"location":1,"type":11,"text":'
            + match.group(1)
            + "}]]]"
            == html
        ):
            return True
        if "backupLockTimer" in html:
            return True
        return False

    def __load_new_blackbox_token(self):
        try:
            if self.padre:
                print("Obtaining new blackbox token, please wait...")
            blackbox_token = getNewBlackBoxToken(self)
            assert any(
                c.isupper() for c in blackbox_token
            ), "The token must contain uppercase letters."
            assert any(
                c.islower() for c in blackbox_token
            ), "The token must contain lowercase letters."
            assert any(
                c.isdigit() for c in blackbox_token
            ), "The token must contain digits."
            self.blackbox = blackbox_token
        except Exception as e:
            self.logger.error("Failed to obtain new blackbox token from API: ", exc_info=True)
            if not self.padre: # only exit if running in a child process because user won't be looking at the console to provide the cookies
                sys.exit('Failed to regenerate blackbox token')
            self.blackbox = 'tra:JVqc1fosb5TG-E2h5Ak7bZLEB23OOq0SN2ms0QM1fOFErx5DdafM_kFmmMoXeNsAMmKxBClbi-MIOmyRwwYrXY_VPrAVe-pih7nrEEKFqtwOVcQzmgZrkMLyO6kMOl-Rw-gaXZG26CtcgbP2G01_5FJ_xggtX5G26CtQgrToToO2GX3eD0eq3D6fADCTw_YrX8EjWL73WIvwU7UaT7QYSXvcPnLWDnOk1TeY0Ac6nc81ZpfI_i9hkcMlXpLH7B5QdafqD0FzuimY_2vQ9SdXoA5xn8T2Jk6VBHPaRqvU-Stur_1EkNX6LFyEyzqpEHzhBjh7oNICWM05pAVzmMr6K1mMuuoPQXGZ7GPMMqb5YcImi_0iVITILaMMb9T5K1uD1kutJ4z-bZa77R1Fde0dTX2t8CBkqdL7IFKVuuwcb-ZPtSl85EWpDoCl1wdr3Ua8IZO84RNFapzfBDZonAEyZMf_ZZX2WLvyI4e58CKEue1Ti8H6XsEih7gZSYDjGlGHt-kiVIS981aIuBx-4hR1rOBFqNwUet8XTX6zFjttn8T2OV6QwiOIwPgxZpv_M2fI-y5gkPIqW4_EKovA8SVdjfMjicHxI1uQwvpglcX-M5b6MWmc0TRs0QNmy_9hmvswZp3RNm6TxfccTpHC9CxcgbP2LV-PtOYpW4-05ilOgLIXeKoNQHncEXbZOm6m3kCk20Cl1TZnywJlygBil839LZHyJ43A9leO8SmK6yNcviKH6RlQiL73J43yKmOYyC1jiLrsEUOGq90PQ3iu5BxTuOxRgbHjHE1_tBdHqeATeNxCda7fRabWCG2iBDx1pdtApts9oNM0l8krXpL3Wb73KoztHVG05RhIqc4AMleJzPEjVYnvJFe6Hn-w6Et930Ch0TRkl8wAYsT5X5j5LJH0VrvwVbnqHH3fE3evFEV22DlxqNs-cNYHOGmf0AIyZMb_M2iNv_EWSIuw4hRIgbPrHoO05BV6quJDd63gQqjcDHCp2j-kBmug0Qk_ctM0bNIIOG6k3BF0rBFGf7cbf7EWerLrJInB-ChelPotUoS22w1QgbPnFUV5rOAXTH616htRgbjsEUOGq90Pc6beQXKl2AhpzwU5m9ADaZwCZpvUBDWW9ylav_YsXI_G9luLw_Qmh7_0KI7G_DNqotsUSn7iRaraEUN4rxRNr9QGOF2P0gQ1bp7SAjZrosf5PGGTxfcnWY266h1KfLUJOnGWyQo6bpPGBztvndQFPJa77R9EdrneEEKw4E2CuyWX0D-pF4sDdKcMP3Op3haN-W6gC3GWyPofUZTF_DBmi70AJVeJuuoYSX6j1QcsXqHG-Cp4vSBYuwJcofUmng5fzDeeAESazTGr7TiR-mrgRInbH3jhL2O6I1i9D3rqQpDoOoztRbTpS7kdg-VSh9E0eLnuP4LQA0-U-yBSmPpo3BBdpfpvxzJz3y2EuC6R_mWdAmu5K4XKOnzQSn_EF2qs9UOv5EuY6xxu0RdhyBxkqhF14i9luCVeuAx1ti589juzAXvIQI_TOrMCVrkxVoi63xFUeavdKpkTfOhUtdoMUoe15Qo8bJThQqUOfPBf0jpfktT5K1ukEobrV3yu3iuM7xRGdsUYPW-f9xxOfq_fDT5zmMsNMmSUBnyh1BVGdq_dDTZbjb0Eacw3psv9Q3Wl1gY2Z5fI7R9Plf5w1TuqIkd5v_AhWIa22w0_ZJbZ_jBilMT2KleHuucZUqbXDjNmp9cLMGOk2Aw6aprKJEl7rdIER7UqlgInXKA'
            print(f'{bcolors.RED}[ERROR]{bcolors.ENDC} Failed to obtain new blackbox token from API: ' + str(e)) # using expired fallback token here so that user can insert cookie manually since blackbox generation failed at this point
            print('Please report this issue to developers on github or the discord server!!')
            print('Using fallback blackbox token: ' + self.blackbox)
            enter()

    def __login(self, retries=0):
        if not self.logged:
            banner()

            self.mail = read(msg="Mail:")

            if len(config.predetermined_input) != 0:
                self.password = config.predetermined_input.pop(0)
            else:
                self.password = getpass.getpass("Password:")

            banner()

        #choose one user agent from user_agents list based on provided mail
        self.user_agent = user_agents[sum(ord(c) for c in self.mail) % len(user_agents)]

        self.s = requests.Session()
        self.cipher = AESCipher(self.mail, self.password)
        self.logger.info("__login()")

        # test to see if the lobby cookie in the session file is valid, this will save time on login and will reduce use of blackbox token
        sessionData = self.getSessionData()
        if "shared" in sessionData and "lobby" in sessionData["shared"]:
            cookie_obj = requests.cookies.create_cookie(
                domain=".gameforge.com",
                name="gf-token-production",
                value=sessionData["shared"]["lobby"]["gf-token-production"],
            )
            self.s.cookies.set_cookie(cookie_obj)

        if not self.__test_lobby_cookie():

            self.logger.warning("Getting new lobby cookie")
            self.__load_new_blackbox_token()

            # get gameEnvironmentId and platformGameId
            self.headers = {
                "Host": "lobby.ikariam.gameforge.com",
                "User-Agent": self.user_agent,
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate",
                "DNT": "1",
                "Connection": "close",
                "Referer": "https://lobby.ikariam.gameforge.com/",
            }
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            r = self.s.get(
                "https://lobby.ikariam.gameforge.com/config/configuration.js"
            )

            js = r.text
            gameEnvironmentId = re.search(r'"gameEnvironmentId":"(.*?)"', js)
            if gameEnvironmentId is None:
                sys.exit("gameEnvironmentId not found")
            gameEnvironmentId = gameEnvironmentId.group(1)
            platformGameId = re.search(r'"platformGameId":"(.*?)"', js)
            if platformGameId is None:
                sys.exit("platformGameId not found")
            platformGameId = platformGameId.group(1)

            # get __cfduid cookie
            self.headers = {
                "User-Agent": self.user_agent,
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate",
                "DNT": "1",
                "Connection": "close",
                "Referer": "https://lobby.ikariam.gameforge.com/",
            }
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            r = self.s.get("https://gameforge.com/js/connect.js")
            html = r.text
            captcha = re.search(r"Attention Required", html)
            if captcha is not None:
                sys.exit("Captcha error!")

            # update __cfduid cookie
            self.headers = {
                "User-Agent": self.user_agent,
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate",
                "Referer": "https://lobby.ikariam.gameforge.com/",
                "Origin": "https://lobby.ikariam.gameforge.com",
                "DNT": "1",
                "Connection": "close",
            }
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            r = self.s.get("https://gameforge.com/config")

            __fp_eval_id_1 = self.__fp_eval_id()
            __fp_eval_id_2 = self.__fp_eval_id()
            try:
                # get pc_idt cookie
                self.headers = {
                    "Host": "pixelzirkus.gameforge.com",
                    "User-Agent": self.user_agent,
                    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                    "Accept-Language": "en-US,en;q=0.5",
                    "Accept-Encoding": "gzip, deflate",
                    "Content-Type": "application/x-www-form-urlencoded",
                    "Origin": "https://lobby.ikariam.gameforge.com",
                    "DNT": "1",
                    "Connection": "close",
                    "Referer": "https://lobby.ikariam.gameforge.com/",
                    "Upgrade-Insecure-Requests": "1",
                }
                self.s.headers.clear()
                self.s.headers.update(self.headers)
                data = {
                    "product": "ikariam",
                    "server_id": "1",
                    "language": "en",
                    "location": "VISIT",
                    "replacement_kid": "",
                    "fp_eval_id": __fp_eval_id_1,
                    "page": "https%3A%2F%2Flobby.ikariam.gameforge.com%2F",
                    "referrer": "",
                    "fingerprint": "2175408712",
                    "fp_exec_time": "1.00",
                }
                r = self.s.post(
                    "https://pixelzirkus.gameforge.com/do/simple", data=data
                )

                # update pc_idt cookie
                self.headers = {
                    "Host": "pixelzirkus.gameforge.com",
                    "User-Agent": self.user_agent,
                    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
                    "Accept-Language": "en-US,en;q=0.5",
                    "Accept-Encoding": "gzip, deflate",
                    "Content-Type": "application/x-www-form-urlencoded",
                    "Origin": "https://lobby.ikariam.gameforge.com",
                    "DNT": "1",
                    "Connection": "close",
                    "Referer": "https://lobby.ikariam.gameforge.com/",
                    "Upgrade-Insecure-Requests": "1",
                }
                self.s.headers.clear()
                self.s.headers.update(self.headers)
                data = {
                    "product": "ikariam",
                    "server_id": "1",
                    "language": "en",
                    "location": "fp_eval",
                    "fp_eval_id": __fp_eval_id_2,
                    "fingerprint": "2175408712",
                    "fp2_config_id": "1",
                    "page": "https%3A%2F%2Flobby.ikariam.gameforge.com%2F",
                    "referrer": "",
                    "fp2_value": "921af958be7cf2f76db1e448c8a5d89d",
                    "fp2_exec_time": "96.00",
                }
                r = self.s.post(
                    "https://pixelzirkus.gameforge.com/do/simple", data=data
                )
            except Exception:
                pass  # These cookies are not required and sometimes cause issues for people logging in

            # options req (not really needed)
            self.headers = {
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate, br",
                "Access-Control-Request-Headers": "content-type,tnt-installation-id",
                "Access-Control-Request-Method": "POST",
                "Origin": "https://lobby.ikariam.gameforge.com",
                "Referer": "https://lobby.ikariam.gameforge.com/",
                "Sec-Fetch-Dest": "empty",
                "Sec-Fetch-Mode": "no-cors",
                "Sec-Fetch-Site": "same-site",
                "TE": "trailers",
                "User-Agent": self.user_agent,
            }
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            r = self.s.options("https://gameforge.com/api/v1/auth/thin/sessions")

            # send creds
            self.headers = {
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate, br",
                "Access-Control-Request-Headers": "content-type,tnt-installation-id",
                "Access-Control-Request-Method": "POST",
                "Origin": "https://lobby.ikariam.gameforge.com",
                "Referer": "https://lobby.ikariam.gameforge.com/",
                "Sec-Fetch-Dest": "empty",
                "Sec-Fetch-Mode": "no-cors",
                "Sec-Fetch-Site": "same-site",
                "TE": "trailers",
                "TNT-Installation-Id": "",
                "User-Agent": self.user_agent,
            }
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            data = {
                "identity": self.mail,
                "password": self.password,
                "locale": "en-GB",
                "gfLang": "en",
                "gameId": platformGameId,
                "gameEnvironmentId": gameEnvironmentId,
                "blackbox": self.blackbox,
            }
            r = self.s.post(
                "https://spark-web.gameforge.com/api/v2/authProviders/mauth/sessions", json=data
            )

            # MFA / 2FA Check. If the server responds with 409, it means 2FA is required.
            if r.status_code == 409 and 'OTP_REQUIRED' in r.text:
                if self.padre:
                    print("Two-factor authentication (2FA) is required.")
                    mfa_code = read(msg="Enter your 2FA code: ")
                else:
                    self.logger.error("2FA is required, but it cannot be requested in a child process.")
                    sys.exit("Login failure: 2FA is required in a non-interactive process.")

                # Add the OTP code to the original data and send the request again
                # to the same endpoint.
                data['otpCode'] = mfa_code

                r = self.s.post(
                    "https://spark-web.gameforge.com/api/v2/authProviders/mauth/sessions", json=data
                )

            if "gf-challenge-id" in r.headers and 'token' not in r.text:
                while True:
                    self.headers = {
                        "Accept": "*/*",
                        "Accept-Language": "en-US,en;q=0.5",
                        "Accept-Encoding": "gzip, deflate, br",
                        "Access-Control-Request-Headers": "content-type,tnt-installation-id",
                        "Access-Control-Request-Method": "POST",
                        "Origin": "https://lobby.ikariam.gameforge.com",
                        "Referer": "https://lobby.ikariam.gameforge.com/",
                        "Sec-Fetch-Dest": "empty",
                        "Sec-Fetch-Mode": "no-cors",
                        "Sec-Fetch-Site": "same-site",
                        "TE": "trailers",
                        "TNT-Installation-Id": "",
                        "User-Agent": self.user_agent,
                    }
                    self.s.headers.clear()
                    self.s.headers.update(self.headers)
                    data = {
                            "identity": self.mail,
                            "password": self.password,
                            "locale": "en-GB",
                            "gfLang": "en",
                            "gameId": platformGameId,
                            "gameEnvironmentId": gameEnvironmentId,
                            "blackbox": self.blackbox,
                        }
                    r = self.s.post(
                        "https://spark-web.gameforge.com/api/v2/authProviders/mauth/sessions", json=data
                    )

                    challenge_id = r.headers["gf-challenge-id"].split(";")[0]
                    self.headers = {
                        "accept": "*/*",
                        "accept-encoding": "gzip, deflate, br",
                        "accept-language": "en-GB,el;q=0.9",
                        "dnt": "1",
                        "origin": "https://lobby.ikariam.gameforge.com",
                        "referer": "https://lobby.ikariam.gameforge.com/",
                        "sec-fetch-dest": "empty",
                        "sec-fetch-mode": "cors",
                        "sec-fetch-site": "same-site",
                        "user-agent": self.user_agent,
                    }
                    self.s.headers.clear()
                    self.s.headers.update(self.headers)
                    request1 = self.s.get(
                        "https://challenge.gameforge.com/challenge/{}".format(
                            challenge_id
                        )
                    )
                    request2 = self.s.get(
                        "https://image-drop-challenge.gameforge.com/index.js"
                    )
                    try:
                        request3 = self.s.post(
                            "https://pixelzirkus.gameforge.com/do2/simple"
                        )
                    except Exception as e:
                        pass

                    captcha_time = self.s.get(
                        "https://image-drop-challenge.gameforge.com/challenge/{}/en-GB".format(
                            challenge_id
                        )
                    ).json()["lastUpdated"]
                    text_image = self.s.get(
                        "https://image-drop-challenge.gameforge.com/challenge/{}/en-GB/text?{}".format(
                            challenge_id, captcha_time
                        )
                    ).content
                    drag_icons = self.s.get(
                        "https://image-drop-challenge.gameforge.com/challenge/{}/en-GB/drag-icons?{}".format(
                            challenge_id, captcha_time
                        )
                    ).content
                    drop_target = self.s.get(
                        "https://image-drop-challenge.gameforge.com/challenge/{}/en-GB/drop-target?{}".format(
                            challenge_id, captcha_time
                        )
                    ).content
                    data = {}
                    try:
                        captcha = str(getInteractiveCaptchaSolution(self, text_image, drag_icons))
                        if not captcha.isnumeric():
                            raise Exception(
                                "Failed to resolve interactive captcha automatically. Server returned bad data: {}".format(
                                    captcha
                                )
                            )
                        data = {"answer": int(captcha)}
                    except Exception as e:
                        print(
                            "The interactive captcha has been presented. Automatic captcha resolution failed because: {}".format(
                                str(e)
                            )
                        )
                        print("Do you want to solve it via Telegram? (Y/n)")
                        config.predetermined_input[:] = (
                            []
                        )  # Unholy way to clear a ListProxy object
                        answer = read(values=["y", "Y", "n", "N"], default="y")
                        if answer.lower() == "n":
                            sys.exit("Captcha error! (Interactive)")

                        sendToBot(self, "", Photo=text_image)
                        sendToBot(
                            self,
                            "Please send the number of the correct image (1, 2, 3 or 4)",
                            Photo=drag_icons,
                        )
                        print("Check your Telegram and do it fast. The captcha expires quickly")
                        captcha_time = time.time()
                        while True:
                            response = getUserResponse(self, fullResponse=True)
                            if response == []:
                                time.sleep(5)
                                continue
                            response = response[-1]
                            if response["date"] < captcha_time:
                                time.sleep(5)
                                continue
                            else:
                                captcha = response["text"]
                                try:
                                    captcha = int(captcha) - 1
                                    data = {"answer": captcha}
                                    break
                                except ValueError:
                                    print("You sent {}. Please send only a number (1, 2, 3 or 4)".format(captcha))
                                    time.sleep(5)
                                    continue
                            time.sleep(5)
                    captcha_sent = self.s.post(
                        "https://image-drop-challenge.gameforge.com/challenge/{}/en-GB".format(
                            challenge_id
                        ),
                        json=data,
                    ).json()
                    if captcha_sent["status"] == "solved":
                        self.headers = {
                            "Accept": "*/*",
                            "Accept-Language": "en-US,en;q=0.5",
                            "Accept-Encoding": "gzip, deflate, br",
                            "Access-Control-Request-Headers": "content-type,tnt-installation-id",
                            "Access-Control-Request-Method": "POST",
                            "Origin": "https://lobby.ikariam.gameforge.com",
                            "Referer": "https://lobby.ikariam.gameforge.com/",
                            "Sec-Fetch-Dest": "empty",
                            "Sec-Fetch-Mode": "no-cors",
                            "Sec-Fetch-Site": "same-site",
                            "TE": "trailers",
                            "TNT-Installation-Id": "",
                            "Gf-Challenge-Id": challenge_id,
                            "User-Agent": self.user_agent,
                        }
                        self.s.headers.clear()
                        self.s.headers.update(self.headers)
                        data = {
                            "identity": self.mail,
                            "password": self.password,
                            "locale": "en-GB",
                            "gfLang": "en",
                            "gameId": platformGameId,
                            "gameEnvironmentId": gameEnvironmentId,
                            "blackbox": self.blackbox,
                        }
                        r = self.s.post(
                            "https://spark-web.gameforge.com/api/v2/authProviders/mauth/sessions", json=data
                        )
                        if "gf-challenge-id" in r.headers:
                            self.logger.error("Failed to solve interactive captcha!")
                            print("Failed to solve interactive captcha, trying again!")
                            continue
                        else:
                            break

            if 'token' not in r.text:
                print("Failed to log in...")
                print(f"Expected to get token in response to login request but instead got code {r.status_code} and body {r.text}")
                print(
                    "Log into the lobby via browser and then press CTRL + SHIFT + J to open up the javascript console"
                )
                print(
                    "If you can not open the console using CTRL + SHIFT + J then press F12 to open Dev Tools"
                )
                print(
                    'In the dev tools there should be a tab called "Console". Press this tab.'
                )
                print("Paste in the script below and press enter")
                print(
                    "document.cookie.split(';').forEach(x => {if (x.includes('production')) console.log(x)})"
                )

                auth_token = read(msg="\nEnter gf-token-production manually:").split(
                    "="
                )[-1]
                cookie_obj = requests.cookies.create_cookie(
                    domain=".gameforge.com",
                    name="gf-token-production",
                    value=auth_token,
                )
                self.s.cookies.set_cookie(cookie_obj)
                if not self.__test_lobby_cookie():
                    sys.exit("Wrong email or password\n")
            else:
                # get the authentication token and set the cookie
                ses_json = json.loads(r.text, strict=False)
                auth_token = ses_json["token"]
                cookie_obj = requests.cookies.create_cookie(
                    domain=".gameforge.com",
                    name="gf-token-production",
                    value=auth_token,
                )
                self.s.cookies.set_cookie(cookie_obj)

            # set the lobby cookie in shared for all world server accounts

            lobby_data = dict()
            lobby_data["lobby"] = dict()
            lobby_data["lobby"]["gf-token-production"] = auth_token
            self.setSessionData(lobby_data, shared=True)
        else:
            self.logger.info("Using old lobby cookie")

        # get accounts
        self.headers = {
            "Host": "lobby.ikariam.gameforge.com",
            "User-Agent": self.user_agent,
            "Accept": "application/json",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate",
            "Referer": "https://lobby.ikariam.gameforge.com/es_AR/hub",
            "Authorization": "Bearer {}".format(self.s.cookies["gf-token-production"]),
            "DNT": "1",
            "Connection": "close",
        }
        self.s.headers.clear()
        self.s.headers.update(self.headers)
        r = self.s.get("https://lobby.ikariam.gameforge.com/api/users/me/accounts")
        accounts = json.loads(r.text, strict=False)

        # get servers
        self.headers = {
            "Host": "lobby.ikariam.gameforge.com",
            "User-Agent": self.user_agent,
            "Accept": "application/json",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate",
            "Referer": "https://lobby.ikariam.gameforge.com/es_AR/hub",
            "Authorization": "Bearer {}".format(self.s.cookies["gf-token-production"]),
            "DNT": "1",
            "Connection": "close",
        }
        self.s.headers.clear()
        self.s.headers.update(self.headers)
        r = self.s.get("https://lobby.ikariam.gameforge.com/api/servers")
        servers = json.loads(r.text, strict=False)

        if not self.logged:

            if (
                len([account for account in accounts if account["blocked"] is False])
                == 1
            ):
                self.account = [
                    account for account in accounts if account["blocked"] is False
                ][0]
            else:
                print("With which account do you want to log in?\n")

                max_name = max(
                    [
                        len(account["name"])
                        for account in accounts
                        if account["blocked"] is False
                    ]
                )
                i = 0
                for account in [
                    account for account in accounts if account["blocked"] is False
                ]:
                    server = account["server"]["language"]
                    mundo = account["server"]["number"]
                    account_group = account["accountGroup"]
                    server_lang = None
                    world, server_lang = [
                        (srv["name"], srv["language"])
                        for srv in servers
                        if srv["accountGroup"] == account_group
                    ][0]
                    try: lastlogin =  lastloginTimetoString(account["lastLogin"])
                    except: lastlogin = 'Unknown'
                    
                    i += 1
                    pad = " " * (max_name - len(account["name"]))
                    print(
                        "({:d}) {}{} [{} - {} - {}]".format(
                            i, account["name"], pad, lastlogin, server_lang, world
                        )
                    )
                num = read(min=1, max=i)
                self.account = [
                    account for account in accounts if account["blocked"] is False
                ][num - 1]
            self.username = self.account["name"]
            self.login_servidor = self.account["server"]["language"]
            self.account_group = self.account["accountGroup"]
            self.mundo = str(self.account["server"]["number"])

            self.word, self.servidor = [
                (srv["name"], srv["language"])
                for srv in servers
                if srv["accountGroup"] == self.account_group
            ][0]

            config.infoUser = "Server:{}".format(self.servidor)
            config.infoUser += ", World:{}".format(self.word)
            config.infoUser += ", Player:{}".format(self.username)
            banner()

        self.host = "s{}-{}.ikariam.gameforge.com".format(self.mundo, self.servidor)
        self.urlBase = "https://{}/index.php?".format(self.host)

        self.headers = {
            "Host": self.host,
            "User-Agent": self.user_agent,
            "Accept": "*/*",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate, br",
            "Referer": "https://{}".format(self.host),
            "X-Requested-With": "XMLHttpRequest",
            "Origin": "https://{}".format(self.host),
            "DNT": "1",
            "Connection": "keep-alive",
            "Pragma": "no-cache",
            "Cache-Control": "no-cache",
        }

        sessionData = self.getSessionData()

        used_old_cookies = False
        # if there are cookies stored, try to use them
        if "cookies" in sessionData and self.logged is False:
            # create a new temporary session object
            old_s = requests.Session()
            # set the headers
            old_s.headers.clear()
            old_s.headers.update(self.headers)
            # set the cookies to test
            cookie_dict = sessionData["cookies"]
            requests.cookies.cookiejar_from_dict(
                cookie_dict, cookiejar=old_s.cookies, overwrite=True
            )
            self.__update_proxy(obj=old_s, sessionData=sessionData)
            try:
                # make a request to check the connection
                html = old_s.get(self.urlBase, verify=config.do_ssl_verify).text
            except Exception:
                self.__proxy_error()

            cookies_are_valid = self.__isExpired(html) is False
            if cookies_are_valid:
                self.logger.info("using old cookies")
                used_old_cookies = True
                # assign the old cookies to the session object
                requests.cookies.cookiejar_from_dict(
                    cookie_dict, cookiejar=self.s.cookies, overwrite=True
                )
                # set the proxy
                self.__update_proxy(sessionData=sessionData)
                # set the headers
                self.s.headers.clear()
                self.s.headers.update(self.headers)

        # login as normal and get new cookies
        if used_old_cookies is False:
            self.__load_new_blackbox_token()
            self.logger.warning("using new cookies")
            self.headers = {
                "authority": "lobby.ikariam.gameforge.com",
                "method": "POST",
                "path": "/api/users/me/loginLink",
                "scheme": "https",
                "accept": "application/json",
                "accept-encoding": "gzip, deflate, br",
                "accept-language": "en-US,en;q=0.9",
                "authorization": "Bearer " + self.s.cookies["gf-token-production"],
                "content-type": "application/json",
                "origin": "https://lobby.ikariam.gameforge.com",
                "referer": "https://lobby.ikariam.gameforge.com/en_GB/accounts",
                "user-agent": self.user_agent,
            }
            self.s.headers.clear()
            self.s.headers.update(self.headers)
            data = {
                "server": {"language": self.login_servidor, "number": self.mundo},
                "clickedButton": "account_list",
                "id": self.account["id"],
                "blackbox": self.blackbox,
            }
            resp = self.s.post(
                "https://lobby.ikariam.gameforge.com/api/users/me/loginLink", json=data
            )
            respJson = json.loads(resp.text)
            skipGetCookie = False
            if "url" not in respJson:
                if retries > 0:
                    return self.__login(retries - 1)
                else:  # 403 is for bad user/pass and 400 is bad blackbox token?
                    msg = (
                        "Login Error: "
                        + str(resp.status_code)
                        + " "
                        + str(resp.reason)
                        + " "
                        + str(resp.text)
                    )
                    self.logger.error(msg)
                    if self.padre:
                        print(msg)
                        print(
                            "Failed to log in... Do you want to provide the cookie manually? (Y|N): "
                        )
                        choice = read(values=["y", "Y", "n", "N"], empty=False)
                        if choice in ["n", "N"]:
                            sys.exit(msg)
                        while True:
                            print("• Log into the account via browser")
                            print("• Load your city view")
                            print("• Press F12 to open up the dev tools")
                            print(
                                '• In the dev tools click the tab "Application" if on Chrome or "Storage" if on Firefox'
                            )
                            print(
                                '• Within this window, there should be a dropdown menu called "Cookies" on the far left'
                            )
                            print(
                                '• Select the row "https://s[nubmer]-[region].ikariam.gameforge.com"'
                            )
                            print(
                                '• Look in the table on the right for the entry named "ikariam"'
                            )
                            print("• Copy its value and paste it just below")

                            ikariam_cookie = read(
                                msg="\nEnter ikariam cookie manually: "
                            ).split("=")[-1]
                            cookie_obj = requests.cookies.create_cookie(
                                domain=self.host, name="ikariam", value=ikariam_cookie
                            )
                            self.s.cookies.set_cookie(cookie_obj)
                            try:
                                # make a request to check the connection
                                html = self.s.get(
                                    self.urlBase, verify=config.do_ssl_verify
                                ).text
                            except Exception:
                                self.__proxy_error()
                            skipGetCookie = cookies_are_valid = (
                                self.__isExpired(html) is False
                            )
                            if not cookies_are_valid:
                                print(
                                    "This cookie is expired. Do you want to try again? (Y|N): "
                                )
                                choice = read(values=["y", "Y", "n", "N"], empty=False)
                                if choice in ["n", "N"]:
                                    sys.exit(msg)
                                continue
                            # TODO check if account is actually the one associated with this email / pass
                            break
                    else:
                        self.logger.error("I wanted to ask user for ikariam cookie, but he wasn't looking")
                        sys.exit(msg)

            if not skipGetCookie:
                url = respJson["url"]
                match = re.search(
                    r"https://s\d+-\w{2}\.ikariam\.gameforge\.com/index\.php\?", url
                )
                if match is None:
                    sys.exit("Error")

                # set the headers
                self.s.headers.clear()
                self.s.headers.update(self.headers)

                # set the proxy
                self.__update_proxy(sessionData=sessionData)

                # use the new cookies instead, invalidate the old ones
                try:
                    html = self.s.get(url, verify=config.do_ssl_verify).text
                except Exception:
                    self.__proxy_error()

        if self.__isInVacation(html):
            msg = "The account went into vacation mode"
            if self.padre:
                print(msg)
            else:
                sendToBot(self, msg)
            os._exit(0)
        if self.__isExpired(html):
            if retries > 0:
                return self.__login(retries - 1)
            if self.padre:
                msg = "Login error."
                print(msg)
                os._exit(0)
            raise Exception("Couldn't log in")

        if not used_old_cookies:
            self.__saveNewCookies()

        self.logged = True

    def __backoff(self):
        self.logger.info("__backoff()")
        if self.padre is False:
            time.sleep(5 * random.randint(0, 10))

    def __sessionExpired(self):
        self.logger.info("__sessionExpired()")
        self.__backoff()

        sessionData = self.getSessionData()

        try:
            if self.s.cookies["PHPSESSID"] != sessionData["cookies"]["PHPSESSID"]:
                self.__getCookie(sessionData)
            else:
                try:
                    self.__login(3)
                except Exception:
                    self.__sessionExpired()
        except KeyError:
            try:
                self.__login(3)
            except Exception:
                self.__sessionExpired()

    def __proxy_error(self):
        sessionData = self.getSessionData()
        if "proxy" not in sessionData or sessionData["proxy"]["set"] is False:
            sys.exit("network error")
        elif self.padre is True:
            print("There seems to be a problem connecting to ikariam.")
            print("Do you want to disable the proxy? [Y/n]")
            rta = read(values=["y", "Y", "n", "N", ""])
            if rta.lower() == "n":
                sys.exit()
            else:
                sessionData["proxy"]["set"] = False
                self.setSessionData(sessionData)
                print("Proxy disabled, try again.")
                enter()
                sys.exit()
        else:
            msg = "Network error. Consider disabling the proxy."
            sendToBot(self, msg)
            sys.exit()

    def __update_proxy(self, *, obj=None, sessionData=None):
        # set the proxy
        if obj is None:
            obj = self.s
        if sessionData is None:
            sessionData = self.getSessionData()
        if "proxy" in sessionData and sessionData["proxy"]["set"] is True:
            obj.proxies.update(sessionData["proxy"]["conf"])
        else:
            obj.proxies.update({})

    def __checkCookie(self):
        self.logger.info("__checkCookie()")
        sessionData = self.getSessionData()

        try:
            if self.s.cookies["PHPSESSID"] != sessionData["cookies"]["PHPSESSID"]:
                self.__getCookie(sessionData)
        except KeyError:
            try:
                self.__login(3)
            except Exception:
                self.__sessionExpired()

    def __token(self):
        """Generates a valid actionRequest token from the session
        Returns
        -------
        token : str
            a string representing a valid actionRequest token
        """
        html = self.get()
        return re.search(r'actionRequest"?:\s*"(.*?)"', html).group(1)

    def get(
        self, url='', params={}, ignoreExpire=False, noIndex=False, fullResponse=False, noQuery=False, **kwargs
    ):
        """Sends get request to ikariam
        Parameters
        ----------
        url : str
            this string will be appended to the end of the urlBase of the Session object. urlBase will look like: 'https://s(number)-(country).ikariam.gameforge.com/index.php?'
        params : dict
            dictionary containing key-value pairs which represent the parameteres of the get request
        ignoreExpire: bool
            if set to True it will ignore if the current session is expired and will simply return whatever response it gets. If it's set to False, it will make sure that the current session is not expired before sending the get request, if it's expired it will login again
        noIndex : bool
            if set to True it will remove 'index.php' from the end of urlBase before appending url params and sending the get request
        fullResponse : bool
            if set to True it will retrn the full response object instead of the string containing html or json data

        Returns
        -------
        html : str
            response from the server
        """
        self.__checkCookie()
        self.__update_proxy()

        if noIndex:
            url = self.urlBase.replace("index.php", "") + url
        else:
            url = self.urlBase + url
        if noQuery:
            url = url.replace('?','')
        while True:
            try:
                self.requestHistory.append(
                    {
                        "method": "GET",
                        "url": url,
                        "params": params,
                        "payload": None,
                        "proxies": self.s.proxies,
                        "headers": dict(self.s.headers),
                        "response": None,
                    }
                )
                self.logger.debug(f"About to send: {str(self.requestHistory[-1])}")
                response = self.s.get(
                    url, params=params, verify=config.do_ssl_verify, timeout=300, **kwargs
                )
                self.requestHistory[-1]["response"] = {
                    "status": response.status_code,
                    "elapsed": response.elapsed.total_seconds(),
                    "headers": dict(response.headers),
                    "text": response.text,
                }
                html = response.text
                if self.__test_server_maintenace(html):
                    self.logger.warning("Ikariam world backup is in progress, waiting 10 mins.")
                    time.sleep(10 * 60)
                    raise requests.exceptions.ConnectionError  # repeat after 10 minutes
                if ignoreExpire is False:
                    assert self.__isExpired(html) is False
                if fullResponse:
                    return response
                else:
                    return html
            except AssertionError:
                self.__sessionExpired()
            except requests.exceptions.ConnectionError:
                self.logger.warning(f"Connection error occured, retrying in {ConnectionError_wait}s\n{str(params) + ' --> ' + url}")
                time.sleep(ConnectionError_wait)
            except requests.exceptions.Timeout:
                self.logger.warning(f"5 minute timeout occured on request, retrying in {ConnectionError_wait}s\n{str(params) + ' --> ' + url}")
                time.sleep(ConnectionError_wait)

    def post(
        self, url='', payloadPost={}, params={}, ignoreExpire=False, noIndex=False, fullResponse = False, noQuery=False, **kwargs
    ):
        """Sends post request to ikariam
        Parameters
        ----------
        url : str
            this string will be appended to the end of the urlBase of the Session object. urlBase will look like: 'https://s(number)-(country).ikariam.gameforge.com/index.php?'
        payloadPost : dict
            dictionary containing key-value pairs which represent the payload of the post request
        params : dict
            dictionary containing key-value pairs which represent the parameteres of the post request
        ignoreExpire: bool
            if set to True it will ignore if the current session is expired and will simply return whatever response it gets. If it's set to False, it will make sure that the current session is not expired before sending the post request, if it's expired it will login again
        noIndex : bool
            if set to True it will remove 'index.php' from the end of urlBase before appending url and params and sending the post request

        Returns
        -------
        html : str
            response from the server
        """
        url_original = url
        payloadPost_original = payloadPost
        params_original = params
        self.__checkCookie()
        self.__update_proxy()

        # add the request id
        token = self.__token()
        url = url.replace(actionRequest, token)
        if "actionRequest" in payloadPost:
            payloadPost["actionRequest"] = token
        if "actionRequest" in params:
            params["actionRequest"] = token

        if noIndex:
            url = self.urlBase.replace("index.php", "") + url
        else:
            url = self.urlBase + url
        if noQuery:
            url = url.replace('?','')
        while True:
            try:
                self.requestHistory.append(
                    {
                        "method": "POST",
                        "url": url,
                        "params": params,
                        "payload": payloadPost,
                        "proxies": self.s.proxies,
                        "headers": dict(self.s.headers),
                        "response": None,
                    }
                )
                self.logger.debug(f"About to send: {str(self.requestHistory[-1])}")
                response = self.s.post(
                    url,
                    data=payloadPost,
                    params=params,
                    verify=config.do_ssl_verify,
                    timeout=300,
                    **kwargs,
                )
                self.requestHistory[-1]["response"] = {
                    "status": response.status_code,
                    "elapsed": response.elapsed.total_seconds(),
                    "headers": dict(response.headers),
                    "text": response.text,
                }
                resp = response.text
                if self.__test_server_maintenace(resp):
                    self.logger.warning("Ikariam world backup is in progress, waiting 10 mins.")
                    time.sleep(10 * 60)
                    raise requests.exceptions.ConnectionError  # repeat after 10 minutes
                if ignoreExpire is False:
                    assert self.__isExpired(resp) is False
                if "TXT_ERROR_WRONG_REQUEST_ID" in resp:
                    self.logger.warning("got TXT_ERROR_WRONG_REQUEST_ID, bad actionRequest")
                    return self.post(
                        url=url_original,
                        payloadPost=payloadPost_original,
                        params=params_original,
                        ignoreExpire=ignoreExpire,
                        noIndex=noIndex,
                    )
                return resp if not fullResponse else response
            except AssertionError:
                self.__sessionExpired()
            except requests.exceptions.ConnectionError:
                self.logger.warning(f"Connection error occured, retrying in {ConnectionError_wait}s\n{str(params) + ' --> ' + url}")
                time.sleep(ConnectionError_wait)
            except requests.exceptions.Timeout:
                self.logger.warning(f"5 minute timeout occured on request, retrying in {ConnectionError_wait}s\n{str(params) + ' --> ' + url}")
                time.sleep(ConnectionError_wait)

    def logout(self):
        """This function kills the current (chlid) process"""
        self.logger.info("logout()")
        if self.padre is False:
            os._exit(0)

    def setSessionData(self, sessionData, shared=False):
        """Encrypts relevant session data and writes it to the .ikabot file
        Parameters
        ----------
        sessionData : dict
            dictionary containing relevant session data, data is written to file using AESCipher.setSessionData
        shared : bool
            Indicates if the new data should be shared among all accounts asociated with the user-password
        """
        self.cipher.setSessionData(self, sessionData, shared=shared)

    def getSessionData(self):
        """Gets relevant session data from the .ikabot file"""
        return self.cipher.getSessionData(self)


def normal_get(url, params={}):
    """Sends a get request to provided url
    Parameters
    ----------
    url : str
        a string representing the url to which to send the get request
    params : dict
        a dictionary containing key-value pairs which represent the parameters of the get request

    Returns
    -------
    response : requests.Response
        a requests.Response object which represents the webservers response. For more information on requests.Response refer to https://requests.readthedocs.io/en/master/api/#requests.Response
    """
    try:

        return requests.get(url, params=params)

    except requests.exceptions.ConnectionError:
        sys.exit("Internet connection failed")
